"""empty message

Revision ID: 0d44655e35fd
Revises: 8b3f167d0418
Create Date: 2017-08-31 14:14:54.187073

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "0d44655e35fd"
down_revision = "8b3f167d0418"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "individual_importances",
        sa.Column("model_id", sa.Integer(), nullable=False),
        sa.Column("entity_id", sa.BigInteger(), nullable=False),
        sa.Column("as_of_date", sa.DateTime(), nullable=False),
        sa.Column("feature", sa.String(), nullable=False),
        sa.Column("method", sa.String(), nullable=False),
        sa.Column("importance_score", sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(["model_id"], ["results.models.model_id"]),
        sa.PrimaryKeyConstraint(
            "model_id", "entity_id", "as_of_date", "feature", "method"
        ),
        schema="results",
    )
    op.create_table(
        "list_predictions",
        sa.Column("model_id", sa.Integer(), nullable=False),
        sa.Column("entity_id", sa.BigInteger(), nullable=False),
        sa.Column("as_of_date", sa.DateTime(), nullable=False),
        sa.Column("score", sa.Numeric(), nullable=True),
        sa.Column("rank_abs", sa.Integer(), nullable=True),
        sa.Column("rank_pct", sa.Float(), nullable=True),
        sa.Column("matrix_uuid", sa.Text(), nullable=True),
        sa.Column("test_label_window", sa.Interval(), nullable=True),
        sa.ForeignKeyConstraint(["model_id"], ["results.models.model_id"]),
        sa.PrimaryKeyConstraint("model_id", "entity_id", "as_of_date"),
        schema="results",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("list_predictions", schema="results")
    op.drop_table("individual_importances", schema="results")
    # ### end Alembic commands ###



import datetime
import os
import tempfile
from collections import OrderedDict

import boto3
import pandas as pd
import pytest
import yaml
from moto import mock_s3
from numpy.testing import assert_almost_equal
from pandas.testing import assert_frame_equal
from unittest import mock

from triage.component.catwalk.storage import (
    MatrixStore,
    CSVMatrixStore,
    FSStore,
    S3Store,
    ProjectStorage,
    ModelStorageEngine,
)

from tests.utils import CallSpy


class SomeClass:

    def __init__(self, val):
        self.val = val


def test_S3Store():
    with mock_s3():
        client = boto3.client("s3")
        client.create_bucket(Bucket="test_bucket", ACL="public-read-write")
        store = S3Store(f"s3://test_bucket/a_path")
        assert not store.exists()
        store.write("val".encode("utf-8"))
        assert store.exists()
        newVal = store.load()
        assert newVal.decode("utf-8") == "val"
        store.delete()
        assert not store.exists()


@mock_s3
def test_S3Store_large():
    client = boto3.client('s3')
    client.create_bucket(Bucket='test_bucket', ACL='public-read-write')

    store = S3Store('s3://test_bucket/a_path')
    assert not store.exists()

    # NOTE: The issue under test (currently) arises when too large a "part"
    # NOTE: is sent to S3 for upload -- greater than its 5 GiB limit on any
    # NOTE: single upload request.
    #
    # NOTE: Though s3fs uploads file parts as soon as its buffer reaches
    # NOTE: 5+ MiB, it does not ensure that its buffer -- and resulting
    # NOTE: upload "parts" -- remain under this limit (as the result of a
    # NOTE: single "write()").
    #
    # NOTE: Therefore, until s3fs adds handling to ensure it never attempts
    # NOTE: to upload such large payloads, we'll handle this in S3Store,
    # NOTE: by chunking out writes to s3fs.
    #
    # NOTE: This is all not only to explain the raison d'etre of this test,
    # NOTE: but also as context for the following warning: The
    # NOTE: payload we'll attempt to write, below, is far less than 5 GiB!!
    # NOTE: (Attempting to provision a 5 GiB string in RAM just for this
    # NOTE: test would be an ENORMOUS drag on test runs, and a conceivable
    # NOTE: disruption, depending on the test environment's resources.)
    #
    # NOTE: As such, this test *may* fall out of sync with either the code
    # NOTE: that it means to test or with the reality of the S3 API -- even
    # NOTE: to the point of self-invalidation. (But, this should do the
    # NOTE: trick; and, we can always increase the payload size here, or
    # NOTE: otherwise tweak configuration, as necessary.)
    one_mb = 2 ** 20
    payload = b"0" * (10 * one_mb)  # 10MiB text of all zeros

    with CallSpy('botocore.client.BaseClient._make_api_call') as spy:
        store.write(payload)

    call_args = [call[0] for call in spy.calls]
    call_methods = [args[1] for args in call_args]

    assert call_methods == [
        'CreateMultipartUpload',
        'UploadPart',
        'UploadPart',
        'CompleteMultipartUpload',
    ]

    upload_args = call_args[1]
    upload_body = upload_args[2]['Body']

    # NOTE: Why is this a BufferIO rather than the underlying buffer?!
    # NOTE: (Would have expected the result of BufferIO.read() -- str.)
    body_length = len(upload_body.getvalue())
    assert body_length == 5 * one_mb

    assert store.exists()
    assert store.load() == payload

    store.delete()
    assert not store.exists()


def test_FSStore():
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpfile = os.path.join(tmpdir, "tmpfile")
        store = FSStore(tmpfile)
        assert not store.exists()
        store.write("val".encode("utf-8"))
        assert store.exists()
        newVal = store.load()
        assert newVal.decode("utf-8") == "val"
        store.delete()
        assert not store.exists()


def test_ModelStorageEngine_nocaching(project_storage):
    mse = ModelStorageEngine(project_storage)
    mse.write('testobject', 'myhash')
    assert mse.exists('myhash')
    assert mse.load('myhash') == 'testobject'
    assert 'myhash' not in mse.cache


def test_ModelStorageEngine_caching(project_storage):
    mse = ModelStorageEngine(project_storage)
    with mse.cache_models():
        mse.write('testobject', 'myhash')
        with mock.patch.object(mse, "_get_store") as get_store_mock:
            assert mse.load('myhash') == 'testobject'
            assert not get_store_mock.called
        assert 'myhash' in mse.cache
    # when cache_models goes out of scope the cache should be empty
    assert 'myhash' not in mse.cache


DATA_DICT = OrderedDict(
    [
        ("entity_id", [1, 2]),
        ("as_of_date", [datetime.date(2017, 1, 1), datetime.date(2017, 1, 1)]),
        ("k_feature", [0.5, 0.4]),
        ("m_feature", [0.4, 0.5]),
        ("label", [0, 1]),
    ]
)

METADATA = {"label_name": "label"}


def matrix_stores():
    df = pd.DataFrame.from_dict(DATA_DICT).set_index(MatrixStore.indices)

    with tempfile.TemporaryDirectory() as tmpdir:
        project_storage = ProjectStorage(tmpdir)
        tmpcsv = os.path.join(tmpdir, "df.csv.gz")
        tmpyaml = os.path.join(tmpdir, "df.yaml")
        with open(tmpyaml, "w") as outfile:
            yaml.dump(METADATA, outfile, default_flow_style=False)
        df.to_csv(tmpcsv, compression="gzip")
        csv = CSVMatrixStore(project_storage, [], "df")
        # first test with caching
        with csv.cache():
            yield csv
        # with the caching out of scope they will be nuked
        # and this last version will not have any cache
        yield csv


def test_MatrixStore_empty():
    for matrix_store in matrix_stores():
        assert not matrix_store.empty


def test_MatrixStore_metadata():
    for matrix_store in matrix_stores():
        assert matrix_store.metadata == METADATA


def test_MatrixStore_columns():
    for matrix_store in matrix_stores():
        assert matrix_store.columns() == ["k_feature", "m_feature"]


def test_MatrixStore_resort_columns():
    for matrix_store in matrix_stores():
        result = matrix_store.matrix_with_sorted_columns(
            ["m_feature", "k_feature"]
        ).values.tolist()
        expected = [[0.4, 0.5], [0.5, 0.4]]
        assert_almost_equal(expected, result)


def test_MatrixStore_already_sorted_columns():
    for matrix_store in matrix_stores():
        result = matrix_store.matrix_with_sorted_columns(
            ["k_feature", "m_feature"]
        ).values.tolist()
        expected = [[0.5, 0.4], [0.4, 0.5]]
        assert_almost_equal(expected, result)


def test_MatrixStore_sorted_columns_subset():
    with pytest.raises(ValueError):
        for matrix_store in matrix_stores():
            matrix_store.matrix_with_sorted_columns(["m_feature"]).values.tolist()


def test_MatrixStore_sorted_columns_superset():
    with pytest.raises(ValueError):
        for matrix_store in matrix_stores():
            matrix_store.matrix_with_sorted_columns(
                ["k_feature", "l_feature", "m_feature"]
            ).values.tolist()


def test_MatrixStore_sorted_columns_mismatch():
    with pytest.raises(ValueError):
        for matrix_store in matrix_stores():
            matrix_store.matrix_with_sorted_columns(
                ["k_feature", "l_feature"]
            ).values.tolist()


def test_MatrixStore_labels_idempotency():
    for matrix_store in matrix_stores():
        assert matrix_store.labels.tolist() == [0, 1]
        assert matrix_store.labels.tolist() == [0, 1]


def test_MatrixStore_save():
    data = {
        "entity_id": [1, 2],
        "as_of_date": [pd.Timestamp(2017, 1, 1), pd.Timestamp(2017, 1, 1)],
        "feature_one": [0.5, 0.6],
        "feature_two": [0.5, 0.6],
        "label": [1, 0]
    }
    df = pd.DataFrame.from_dict(data)
    labels = df.pop("label")

    for matrix_store in matrix_stores():
        matrix_store.metadata = METADATA

        matrix_store.matrix_label_tuple = df, labels
        matrix_store.save()
        assert_frame_equal(
            matrix_store.design_matrix,
            df
        )


def test_MatrixStore_caching():
    for matrix_store in matrix_stores():
        with matrix_store.cache():
            matrix = matrix_store.design_matrix
            with mock.patch.object(matrix_store, "_load") as load_mock:
                assert_frame_equal(matrix_store.design_matrix, matrix)
                assert not load_mock.called


def test_as_of_dates(project_storage):
    data = {
        "entity_id": [1, 2, 1, 2],
        "feature_one": [0.5, 0.6, 0.5, 0.6],
        "feature_two": [0.5, 0.6, 0.5, 0.6],
        "as_of_date": [
            pd.Timestamp(2016, 1, 1),
            pd.Timestamp(2016, 1, 1),
            pd.Timestamp(2017, 1, 1),
            pd.Timestamp(2017, 1, 1),
        ],
        "label": [1, 0, 1, 0]
    }
    df = pd.DataFrame.from_dict(data)
    matrix_store = CSVMatrixStore(
        project_storage,
        [],
        "test",
        matrix=df,
        metadata={"indices": ["entity_id", "as_of_date"], "label_name": "label"}
    )
    assert matrix_store.as_of_dates == [datetime.date(2016, 1, 1), datetime.date(2017, 1, 1)]


def test_s3_save():
    with mock_s3():
        client = boto3.client("s3")
        client.create_bucket(Bucket="fake-matrix-bucket", ACL="public-read-write")
        for example in matrix_stores():
            if not isinstance(example, CSVMatrixStore):
                continue
            project_storage = ProjectStorage("s3://fake-matrix-bucket")

            tosave = CSVMatrixStore(project_storage, [], "test")
            tosave.metadata = example.metadata
            tosave.matrix_label_tuple = example.matrix_label_tuple
            tosave.save()

            tocheck = CSVMatrixStore(project_storage, [], "test")
            assert tocheck.metadata == example.metadata
            assert tocheck.design_matrix.to_dict() == example.design_matrix.to_dict()



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import textwrap
from triage.database_reflection import table_row_count, table_exists

DEFAULT_LABEL_NAME = "outcome"


class LabelGeneratorNoOp:
    def generate_all_labels(self, labels_table, as_of_dates, label_timespans):
        logger.warning(
            "No label configuration is available, so no labels will be created"
        )

    def generate(self, start_date, label_timespan, labels_table):
        logger.warning(
            "No label configuration is available, so no labels will be created"
        )

    def clean_up(self, labels_table_name):
        pass


class LabelGenerator:
    def __init__(self, db_engine, query, label_name=None, replace=True):
        self.db_engine = db_engine
        self.replace = replace
        # query is expected to select a number of entity ids
        # and an outcome for each given an as-of-date
        self.query = query
        self.label_name = label_name or DEFAULT_LABEL_NAME

    def _create_labels_table(self, labels_table_name):
        if self.replace or not table_exists(labels_table_name, self.db_engine):
            self.db_engine.execute(f"drop table if exists {labels_table_name}")
            self.db_engine.execute(
                f"""
                create table {labels_table_name} (
                entity_id int,
                as_of_date date,
                label_timespan interval,
                label_name varchar,
                label_type varchar,
                label smallint
                )"""
            )
        else:
            logger.notice(f"Not dropping and recreating {labels_table_name} table because "
                          f"replace flag was set to False and table was found to exist")

    def generate_all_labels(self, labels_table, as_of_dates, label_timespans):
        self._create_labels_table(labels_table)
        logger.spam(f"Creating labels for {len(as_of_dates)} as of dates and {len(label_timespans)} label timespans")
        for as_of_date in as_of_dates:
            for label_timespan in label_timespans:
                if not self.replace:
                    logger.spam(f"Looking for existing labels for as of date {as_of_date} and label timespan {label_timespan}")
                    any_existing_labels = list(self.db_engine.execute(
                        f"""select 1 from {labels_table}
                        where as_of_date = '{as_of_date}'
                        and label_timespan = '{label_timespan}'::interval
                        and label_name = '{self.label_name}'
                        limit 1
                        """)
                    )
                    if len(any_existing_labels) == 1:
                        logger.spam("Since nonzero existing labels found, skipping")
                        continue

                logger.debug(
                    f"Generating labels for as of date {as_of_date} and label timespan {label_timespan}",
                )
                self.generate(
                    start_date=as_of_date,
                    label_timespan=label_timespan,
                    labels_table=labels_table,
                )

        self.db_engine.execute(
            f"create index on {labels_table} (entity_id, as_of_date)"
        )
        logger.spam("Added index to labels table")

        nrows = table_row_count(labels_table, self.db_engine)

        if nrows == 0:
            logger.warning(f"Done creating labels, but no rows in {labels_table} table!")
            raise ValueError(f"{label_table} is empty!")
        else:
            logger.debug(f"Labels table generated at {labels_table}")
            logger.spam(f"Row count of {labels_table}: nrows")

    def generate(self, start_date, label_timespan, labels_table):
        """Generate labels table using a query

        Parameters
        ----------
        start_date: str
            as of date
        label_timespan: str
            postgresql readable time interval
        labels_table: str
            name of labels table
        """
        # we want to apply the as-of-date and label in the database driver,
        # so replace the user {as_of_date} with the SQL %(as_of_date)

        query_with_db_variables = self.query.format(
            as_of_date=start_date, label_timespan=label_timespan
        )

        full_insert_query = textwrap.dedent(
            f"""
            insert into {labels_table}
            select
                entities_and_outcomes.entity_id,
                '{start_date}' as as_of_date,
                '{label_timespan}'::interval as label_timespan,
                '{self.label_name}' as label_name,
                'binary' as label_type,
                entities_and_outcomes.outcome as label
            from ({query_with_db_variables}) entities_and_outcomes
            """
        )

        logger.spam("Running label insertion query")
        logger.spam(full_insert_query)
        self.db_engine.execute(full_insert_query)

    def clean_up(self, labels_table_name):
        self.db_engine.execute(f"drop table if exists {labels_table_name}")



# -*- coding: utf-8 -*-
"""test_collate

Unit tests for `collate` module.

"""
from triage.component.collate import Aggregate, Aggregation, Categorical

def test_aggregate():
    agg = Aggregate("*", "count", {})
    assert list(map(str, agg.get_columns())) == ["count(*)"]

def test_aggregate_cast():
    agg = Aggregate("*", "count", {}, coltype="REAL")
    assert list(map(str, agg.get_columns())) == ["count(*)::REAL"]

def test_categorical_cast():
    cat = Categorical("c", ['A','B','C'], "sum", {}, coltype="SMALLINT")
    assert list(map(str, cat.get_columns())) == [
        "sum((c = 'A')::INT)::SMALLINT",
        "sum((c = 'B')::INT)::SMALLINT",
        "sum((c = 'C')::INT)::SMALLINT"
    ]

def test_aggregate_when():
    agg = Aggregate("1", "count", {})
    assert list(map(str, agg.get_columns(when="date < '2012-01-01'"))) == [
        "count(1) FILTER (WHERE date < '2012-01-01')"
    ]

def test_aggregate_when_cast():
    agg = Aggregate("", "mode", {}, "x", coltype="SMALLINT")
    assert list(map(str, agg.get_columns(when="date < '2012-01-01'"))) == [
        "mode() WITHIN GROUP (ORDER BY x) FILTER (WHERE date < '2012-01-01')::SMALLINT"
    ]


def test_ordered_aggregate():
    agg = Aggregate("", "mode", {}, "x")
    (expression,) = agg.get_columns()
    assert str(expression) == "mode() WITHIN GROUP (ORDER BY x)"
    assert expression.name == "x_mode"


def test_ordered_aggregate_when():
    agg = Aggregate("", "mode", {}, "x")
    assert list(map(str, agg.get_columns(when="date < '2012-01-01'"))) == [
        "mode() WITHIN GROUP (ORDER BY x) FILTER (WHERE date < '2012-01-01')"
    ]




def test_aggregate_tuple_quantity():
    agg = Aggregate(("x", "y"), "corr", {})
    assert list(map(str, agg.get_columns())) == ["corr(x, y)"]


def test_aggregate_tuple_quantity_when():
    agg = Aggregate(("x", "y"), "corr", {})
    assert list(map(str, agg.get_columns(when="date < '2012-01-01'"))) == [
        "corr(x, y) FILTER (WHERE date < '2012-01-01')"
    ]


def test_aggregate_imputation_lookup():
    agg = Aggregate(
        "a",
        ["avg", "sum"],
        {
            "coltype": "aggregate",
            "avg": {"type": "mean"},
            "sum": {"type": "constant", "value": 3},
            "max": {"type": "zero"},
        },
    )
    assert agg.column_imputation_lookup()["a_avg"]["type"] == "mean"
    assert agg.column_imputation_lookup()["a_avg"]["coltype"] == "aggregate"
    assert agg.column_imputation_lookup()["a_sum"]["type"] == "constant"
    assert agg.column_imputation_lookup()["a_sum"]["value"] == 3
    assert agg.column_imputation_lookup()["a_sum"]["coltype"] == "aggregate"


def test_aggregate_imputation_lookup_all():
    agg = Aggregate(
        "a",
        ["avg", "sum"],
        {
            "coltype": "aggregate",
            "all": {"type": "zero"},
            "sum": {"type": "constant", "value": 3},
            "max": {"type": "mean"},
        },
    )
    assert agg.column_imputation_lookup()["a_avg"]["type"] == "zero"
    assert agg.column_imputation_lookup()["a_avg"]["coltype"] == "aggregate"
    assert agg.column_imputation_lookup()["a_sum"]["type"] == "constant"
    assert agg.column_imputation_lookup()["a_sum"]["value"] == 3
    assert agg.column_imputation_lookup()["a_sum"]["coltype"] == "aggregate"


def test_aggregate_arithmetic():
    n = Aggregate("x", "sum", {})
    d = Aggregate("1", "count", {})
    m = Aggregate("y", "avg", {})

    (e,) = (n / d + m).get_columns(prefix="prefix_")
    assert str(e) == "((sum(x)*1.0 / count(1)) + avg(y))"
    assert e.name == "prefix_x_sum/1_count+y_avg"


def test_aggregate_format_kwargs():
    agg = Aggregate("'{collate_date}' - date", "min", {})
    assert list(
        map(str, agg.get_columns(format_kwargs={"collate_date": "2012-01-01"}))
    ) == ["min('2012-01-01' - date)"]


def test_aggregation_table_name_no_schema():
    # no schema
    assert (
        Aggregation(
            [], from_obj="source", groups=[], state_table="tbl"
        ).get_table_name()
        == '"source_aggregation"'
    )
    assert (
        Aggregation([], from_obj="source", groups=[], state_table="tbl").get_table_name(
            imputed=True
        )
        == '"source_aggregation_imputed"'
    )

    # prefix
    assert (
        Aggregation(
            [], from_obj="source", prefix="mysource", groups=[], state_table="tbl"
        ).get_table_name()
        == '"mysource_aggregation"'
    )
    assert (
        Aggregation(
            [], from_obj="source", prefix="mysource", groups=[], state_table="tbl"
        ).get_table_name(imputed=True)
        == '"mysource_aggregation_imputed"'
    )

    # schema
    assert (
        Aggregation(
            [], from_obj="source", schema="schema", groups=[], state_table="tbl"
        ).get_table_name()
        == '"schema"."source_aggregation"'
    )
    assert (
        Aggregation(
            [], from_obj="source", schema="schema", groups=[], state_table="tbl"
        ).get_table_name(imputed=True)
        == '"schema"."source_aggregation_imputed"'
    )


def test_distinct():
    assert list(map(str, Aggregate("distinct x", "count", {}).get_columns())) == [
        "count(distinct x)"
    ]

    assert list(
        map(
            str,
            Aggregate("distinct x", "count", {}).get_columns(
                when="date < '2012-01-01'"
            ),
        )
    ) == ["count(distinct x) FILTER (WHERE date < '2012-01-01')"]

    assert list(
        map(
            str,
            Aggregate("distinct(x)", "count", {}).get_columns(
                when="date < '2012-01-01'"
            ),
        )
    ) == ["count(distinct (x)) FILTER (WHERE date < '2012-01-01')"]

    assert list(
        map(
            str,
            Aggregate("distinct(x,y)", "count", {}).get_columns(
                when="date < '2012-01-01'"
            ),
        )
    ) == ["count(distinct (x,y)) FILTER (WHERE date < '2012-01-01')"]



import os.path

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from alembic.config import Config
from alembic import script
from alembic import command
from triage import create_engine
from triage.database_reflection import table_exists


from .schema import (
    Base,
    Experiment,
    FeatureImportance,
    IndividualImportance,
    ListPrediction,
    ExperimentMatrix,
    Matrix,
    ExperimentModel,
    ExperimentRun,
    ExperimentRunStatus,
    Model,
    ModelGroup,
    Subset,
    TestEvaluation,
    TrainEvaluation,
    TestPrediction,
    TrainPrediction,
    TestPredictionMetadata,
    TrainPredictionMetadata,
    TrainAequitas,
    TestAequitas
)


__all__ = (
    "Base",
    "Experiment",
    "FeatureImportance",
    "IndividualImportance",
    "ListPrediction",
    "ExperimentMatrix",
    "Matrix",
    "ExperimentModel",
    "ExperimentRun",
    "ExperimentRunStatus",
    "Model",
    "ModelGroup",
    "Subset",
    "TestEvaluation",
    "TrainEvaluation",
    "TestPrediction",
    "TrainPrediction",
    "TestPredictionMetadata",
    "TrainPredictionMetadata",
    "TestAequitas",
    "TrainAequitas",
    "mark_db_as_upgraded",
    "upgrade_db",
)


def _base_alembic_args(db_config_filename=None):
    path = os.path.abspath(__file__)
    dir_path = os.path.dirname(path)
    alembic_ini_path = os.path.join(dir_path, "alembic.ini")
    base = ["-c", alembic_ini_path]
    if db_config_filename:
        base += ["-x", "db_config_file={}".format(db_config_filename)]

    return base


def upgrade_db(db_engine=None, dburl=None, revision="head"):
    if db_engine:
        command.upgrade(alembic_config(dburl=db_engine.url), revision)
    elif dburl:
        command.upgrade(alembic_config(dburl=dburl), revision)
    else:
        raise ValueError("Must pass either a db_config_filehandle or a db_engine or a dburl")


def downgrade_db(db_engine=None, dburl=None, revision="-1"):
    if db_engine:
        command.downgrade(alembic_config(dburl=db_engine.url), revision)
    elif dburl:
        command.downgrade(alembic_config(dburl=dburl), revision)
    else:
        raise ValueError("Must pass either a db_config_filehandle or a db_engine or a dburl")


def stamp_db(revision, dburl):
    command.stamp(alembic_config(dburl=dburl), revision)


def db_history(dburl):
    command.history(alembic_config(dburl=dburl))


def upgrade_if_clean(dburl):
    """Upgrade the database only if the results schema hasn't been created yet.

    Raises: ValueError if the database results schema version does not equal the code's version
    """
    alembic_cfg = alembic_config(dburl)
    engine = create_engine(dburl)
    script_ = script.ScriptDirectory.from_config(alembic_cfg)
    if not table_exists('results_schema_versions', engine):
        logger.info("No results_schema_versions table exists, which means that this installation "
                     "is fresh. Upgrading db.")
        upgrade_db(dburl=dburl)
        return
    with engine.begin() as conn:
        current_revision = conn.execute(
            'select version_num from results_schema_versions limit 1'
        ).scalar()
        logger.debug("Database's triage_metadata schema version is %s", current_revision)
        triage_head = script_.get_current_head()
        logger.debug("Code's triage_metadata schema version is %s", triage_head)
        database_is_ahead = not any(
            migration.revision == current_revision
            for migration in script_.walk_revisions()
        )
        if database_is_ahead:
            raise ValueError(
                f"Your database's results schema version, {current_revision}, is not a known "
                "revision to this version of Triage. Usually, this happens if you use a branch "
                "with a new results schema version and upgrade the database to that version. "
                "To use this version of Triage, you will likely need to check out that branch "
                f"and downgrade to {triage_head}",
            )
        elif current_revision != triage_head:
            raise ValueError(
                f"Your database's results schema revision, {current_revision}, is out of date "
                "for this version of Triage. However, your database can be upgraded to this "
                "revision. If you would like to upgrade your database from the console, and "
                "you've installed Triage, you may execute `triage db upgrade`. "
                "If the `triage` command is unavailable, (because you are running Triage directly "
                " from a repository checkout), then `manage alembic upgrade head`. "
                "The database changes may take a long time on a heavily populated database. "
                "Otherwise, you can also downgrade your Triage version to match your database."
            )


def alembic_config(dburl):
    path = os.path.abspath(__file__)
    dir_path = os.path.dirname(path)
    alembic_ini_path = os.path.join(dir_path, "alembic.ini")
    config = Config(alembic_ini_path)
    alembic_path = os.path.join(dir_path, "alembic")
    config.set_main_option("script_location", alembic_path)
    config.attributes["url"] = dburl
    return config



# -*- coding: utf-8 -*-
"""test_imputations

Unit tests for `collate.imputations` module.

"""

from triage.component.collate.imputations import (
    BaseImputation,
    ImputeMean,
    ImputeConstant,
    ImputeZero,
    ImputeZeroNoFlag,
    ImputeNullCategory,
    ImputeBinaryMode,
    ImputeError
)


def test_impute_flag():
    imp = BaseImputation(column="a", coltype="aggregate")
    assert (
        imp.imputed_flag_select_and_alias() == (
            'CASE WHEN "a" IS NULL THEN 1::SMALLINT ELSE 0::SMALLINT END',
            'a_imp'
        )
    )


def test_impute_flag_categorical():
    imp = BaseImputation(column="a", coltype="categorical")
    assert imp.imputed_flag_select_and_alias() == (None, None)


def test_mean_imputation():
    imp = ImputeMean(column="a", coltype="aggregate")
    assert imp.to_sql() == 'COALESCE("a", AVG("a") OVER ()::REAL, 0::REAL) AS "a" '

    imp = ImputeMean(column="a", coltype="aggregate", partitionby="date")
    assert imp.to_sql() == 'COALESCE("a", AVG("a") OVER (PARTITION BY date)::REAL, 0::REAL) AS "a" '

    imp = ImputeMean(column="a", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a", AVG("a") OVER ()::REAL, 0::REAL) AS "a" '

    imp = ImputeMean(column="a", coltype="aggregate", partitionby="date")
    assert imp.to_sql() == 'COALESCE("a", AVG("a") OVER (PARTITION BY date)::REAL, 0::REAL) AS "a" '

    imp = ImputeMean(column="a__NULL_mean", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a__NULL_mean", 1) AS "a__NULL_mean" '


def test_constant_imputation():
    imp = ImputeConstant(column="a", coltype="aggregate", value=3.14)
    assert imp.to_sql() == 'COALESCE("a", 3.14) AS "a" '

    imp = ImputeConstant(column="a_myval_max", coltype="categorical", value="myval")
    assert imp.to_sql() == 'COALESCE("a_myval_max", 1::SMALLINT) AS "a_myval_max" '

    imp = ImputeConstant(column="a_otherval_max", coltype="categorical", value="myval")
    assert imp.to_sql() == 'COALESCE("a_otherval_max", 0::SMALLINT) AS "a_otherval_max" '

    imp = ImputeConstant(column="a__NULL_mean", coltype="categorical", value="myval")
    assert imp.to_sql() == 'COALESCE("a__NULL_mean", 1::SMALLINT) AS "a__NULL_mean" '


def test_impute_zero():
    imp = ImputeZero(column="a", coltype="aggregate")
    assert imp.to_sql() == 'COALESCE("a", 0::REAL) AS "a" '

    imp = ImputeZero(column="a_myval_max", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a_myval_max", 0::SMALLINT) AS "a_myval_max" '

    imp = ImputeZero(column="a_otherval_max", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a_otherval_max", 0::SMALLINT) AS "a_otherval_max" '

    imp = ImputeZero(column="a__NULL_mean", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a__NULL_mean", 1::SMALLINT) AS "a__NULL_mean" '


def test_impute_zero_noflag():
    imp = ImputeZeroNoFlag(column="a", coltype="aggregate")
    assert imp.to_sql() == 'COALESCE("a", 0::SMALLINT) AS "a" '
    assert imp.imputed_flag_select_and_alias() == (None, None)
    assert imp.noflag

    imp = ImputeZeroNoFlag(column="a_myval_max", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a_myval_max", 0::SMALLINT) AS "a_myval_max" '

    imp = ImputeZeroNoFlag(column="a_otherval_max", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a_otherval_max", 0::SMALLINT) AS "a_otherval_max" '

    imp = ImputeZeroNoFlag(column="a__NULL_mean", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a__NULL_mean", 0::SMALLINT) AS "a__NULL_mean" '


def test_impute_null_cat():
    imp = ImputeNullCategory(column="a_myval_max", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a_myval_max", 0::SMALLINT) AS "a_myval_max" '

    imp = ImputeNullCategory(column="a_otherval_max", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a_otherval_max", 0::SMALLINT) AS "a_otherval_max" '

    imp = ImputeNullCategory(column="a__NULL_mean", coltype="categorical")
    assert imp.to_sql() == 'COALESCE("a__NULL_mean", 1::SMALLINT) AS "a__NULL_mean" '

    try:
        imp = ImputeNullCategory(column="a", coltype="aggregate")
        imp.to_sql()
        assert False
    except ValueError:
        assert True
    else:
        assert False


def test_impute_binary_mode():
    imp = ImputeBinaryMode(column="a", coltype="aggregate")
    assert (
        imp.to_sql()
        == 'COALESCE("a", CASE WHEN AVG("a") OVER () > 0.5 THEN 1 ELSE 0 END, 0) AS "a" '
    )

    imp = ImputeBinaryMode(column="a", coltype="aggregate", partitionby="date")
    assert (
        imp.to_sql()
        == 'COALESCE("a", CASE WHEN AVG("a") OVER (PARTITION BY date) > 0.5 '
        'THEN 1 ELSE 0 END, 0) AS "a" '
    )

    try:
        imp = ImputeBinaryMode(column="a", coltype="categorical")
        imp.to_sql()
        assert False
    except ValueError:
        assert True
    else:
        assert False


def test_impute_error():
    try:
        imp = ImputeError(column="a", coltype="aggregate")
        imp.to_sql()
        assert False
    except ValueError:
        assert True
    else:
        assert False



import importlib

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from itertools import permutations
from datetime import datetime
from textwrap import dedent

from sklearn.model_selection import ParameterGrid

from triage.component import architect
from triage.component import catwalk
from triage.component.timechop import Timechop

from triage.util.conf import convert_str_to_relativedelta
from triage.validation_primitives import string_is_tablesafe


class Validator:
    def __init__(self, db_engine=None, strict=True):
        self.db_engine = db_engine
        self.strict = strict

    def run(self, *args, **kwargs):
        try:
            self._run(*args, **kwargs)
        except ValueError as e:
            if self.strict:
                raise ValueError(e)
            else:
                logger.warning(
                    "Validation error hit, not running in strict mode so continuing on: %s",
                    str(e),
                )


class TemporalValidator(Validator):
    def _run(self, temporal_config):
        logger.spam("Validating temporal configuration")
        def dt_from_str(dt_str):
            return datetime.strptime(dt_str, "%Y-%m-%d")

        splits = []
        try:
            chopper = Timechop(
                feature_start_time=dt_from_str(temporal_config["feature_start_time"]),
                feature_end_time=dt_from_str(temporal_config["feature_end_time"]),
                label_start_time=dt_from_str(temporal_config["label_start_time"]),
                label_end_time=dt_from_str(temporal_config["label_end_time"]),
                model_update_frequency=temporal_config["model_update_frequency"],
                training_label_timespans=temporal_config["training_label_timespans"],
                test_label_timespans=temporal_config["test_label_timespans"],
                training_as_of_date_frequencies=temporal_config[
                    "training_as_of_date_frequencies"
                ],
                test_as_of_date_frequencies=temporal_config[
                    "test_as_of_date_frequencies"
                ],
                max_training_histories=temporal_config["max_training_histories"],
                test_durations=temporal_config["test_durations"],
            )
            splits = chopper.chop_time()
        except Exception as e:
            raise ValueError(
                dedent(
                    """
            Section: temporal_config -
            Timechop could not produce temporal splits from config {}.
            Error: {}
            """.format(
                        temporal_config, e
                    )
                )
            )
        for split_num, split in enumerate(splits):
            if len(split["train_matrix"]["as_of_times"]) == 0:
                raise ValueError(
                    dedent(
                        """
                Section: temporal_config -
                Computed split {} has a train matrix with no as_of_times.
                """.format(
                            split
                        )
                    )
                )

            # timechop computes the last time available to train data
            # and stores it in the matrix as 'matrix_info_end_time'
            # but to be more sure, let's double-check by comparing as_of_times
            # in the train and all associated test matrices
            train_max_data_time = max(
                split["train_matrix"]["as_of_times"]
            ) + convert_str_to_relativedelta(
                split["train_matrix"]["training_label_timespan"]
            )

            for test_matrix in split["test_matrices"]:
                if len(test_matrix["as_of_times"]) == 0:
                    raise ValueError(
                        dedent(
                            """
                    Section: temporal_config -
                    Computed split {} has a test matrix with no as_of_times.
                    """.format(
                                split
                            )
                        )
                    )
                overlapping_times = [
                    as_of_time
                    for as_of_time in test_matrix["as_of_times"]
                    if as_of_time < train_max_data_time
                ]
                if overlapping_times:
                    raise ValueError(
                        dedent(
                            """
                    Section: temporal_config -
                    Computed split index {} has a test matrix with as_of_times {}
                    < the maximum train as_of_time + train label timespan.
                    ({}). This is likely an error in timechop. See the
                    experiment's split_definitions[{}] for more information""".format(
                                split_num,
                                overlapping_times,
                                train_max_data_time,
                                split_num,
                            )
                        )
                    )


        logger.debug("Validation of temporal configuration was successful")

class FeatureAggregationsValidator(Validator):
    def _validate_keys(self, aggregation_config):
        logger.spam("Validating feature aggregation keys")
        for key in [
            "from_obj",
            "intervals",
            "groups",
            "knowledge_date_column",
            "prefix",
        ]:
            if key not in aggregation_config:
                raise ValueError(
                    dedent(
                        """
                Section: feature_aggregations -
                '{} required as key: aggregation config: {}""".format(
                            key, aggregation_config
                        )
                    )
                )
                if not string_is_tablesafe(aggregation_config['prefix']):
                    raise ValueError(
                        dedent(
                            f"""Section: feature_aggregations -
                            Feature aggregation prefix should only contain
                            lowercase letters, numbers, and underscores.
                            Aggregation config: {aggregation_config}
                            """
                        )
                    )

        logger.debug("Validation of feature aggregation keys was successful")
    def _validate_aggregates(self, aggregation_config):
        logger.spam("Validating aggregates")
        if (
            "aggregates" not in aggregation_config
            and "categoricals" not in aggregation_config
            and "array_categoricals" not in aggregation_config
        ):
            raise ValueError(
                dedent(
                    """
            Section: feature_aggregations -
            Need either aggregates, categoricals, or array_categoricals
            in {}""".format(
                        aggregation_config
                    )
                )
            )
        logger.debug("Validation of aggregates was successful")

    def _validate_categoricals(self, categoricals):
        logger.spam("Validating categoricals")
        conn = self.db_engine.connect()
        for categorical in categoricals:
            if "choice_query" in categorical and "choices" in categorical:
                raise ValueError(
                    dedent(
                        """
                Section: feature_aggregations -
                Both 'choice_query' and 'choices' specified for {}.
                Please only specify one.""".format(
                            categorical
                        )
                    )
                )
            if not ("choice_query" in categorical or "choices" in categorical):
                raise ValueError(
                    dedent(
                        """
                Section: feature_aggregations -
                Neither 'choice_query' and 'choices' specified for {}.
                Please specify one.""".format(
                            categorical
                        )
                    )
                )
            if "choice_query" in categorical:
                logger.spam("Validating choice query")
                choice_query = categorical["choice_query"]
                try:
                    conn.execute("explain {}".format(choice_query))
                    logger.debug("Validation of choice query was successful")
                except Exception as e:
                    raise ValueError(
                        dedent(
                            """
                    Section: feature_aggregations -
                    choice query does not run.
                    choice query: "{}"
                    Full error: {}""".format(
                                choice_query, e
                            )
                        )
                    )

        logger.debug("Validation of categoricals was successful")
    def _validate_from_obj(self, from_obj):
        conn = self.db_engine.connect()
        logger.spam("Validating from_obj")
        try:
            conn.execute("explain select * from {}".format(from_obj))
            logger.debug("Validation of from_obj was successful")
        except Exception as e:
            raise ValueError(
                dedent(
                    """
                Section: feature_aggregations -
                from_obj query does not run.
                from_obj: "{}"
                Full error: {}""".format(
                        from_obj, e
                    )
                )
            )

    def _validate_time_intervals(self, intervals):
        logger.spam("Validating time intervals")
        for interval in intervals:
            if interval != "all":
                # this function, used elsewhere to break up time intervals,
                # will throw an error if the interval can't be converted to a
                # relativedelta
                try:
                    convert_str_to_relativedelta(interval)
                    logger.debug("Validation of time intervals was successful")
                except Exception as e:
                    raise ValueError(
                        dedent(
                            """
                    Section: feature_aggregations -
                    Time interval is invalid.
                    interval: "{}"
                    Full error: {}""".format(
                                interval, e
                            )
                        )
                    )

    def _validate_groups(self, groups):
        logger.spam("Validating groups")
        if "entity_id" not in groups:
            raise ValueError(
                dedent(
                    """
            Section: feature_aggregations -
            List of groups needs to include 'entity_id'.
            Passed list: {}""".format(
                        groups
                    )
                )
            )
        logger.debug("Validation of groups was successful")

    def _validate_imputation_rule(self, aggregate_type, impute_rule):
        """Validate the imputation rule for a given aggregation type."""
        logger.spam("Validating imputation rule")
        # dictionary of imputation type : required parameters
        valid_imputations = {
            "all": {
                "mean": [],
                "constant": ["value"],
                "zero": [],
                "zero_noflag": [],
                "error": [],
            },
            "aggregates": {"binary_mode": []},
            "categoricals": {"null_category": []},
        }
        valid_imputations["array_categoricals"] = valid_imputations["categoricals"]

        # the valid imputation rules for the specific aggregation type being checked
        valid_types = dict(
            valid_imputations["all"], **valid_imputations[aggregate_type]
        )

        # no imputation rule was specified
        if "type" not in impute_rule.keys():
            raise ValueError(
                dedent(
                    """
            Section: feature_aggregations -
            Imputation type must be specified"""
                )
            )

        # a rule was specified, but not valid for this type of aggregate
        if impute_rule["type"] not in valid_types.keys():
            raise ValueError(
                dedent(
                    """
            Section: feature_aggregations -
            Invalid imputation type %s for %s"""
                    % (impute_rule["type"], aggregate_type)
                )
            )

        # check that all required parameters exist in the keys of the imputation rule
        required_params = valid_types[impute_rule["type"]]
        for param in required_params:
            if param not in impute_rule.keys():
                raise ValueError(
                    dedent(
                        """
                Section: feature_aggregations -
                Missing param %s for %s"""
                        % (param, impute_rule["type"])
                    )
                )
        logger.debug("Validation of imputation rule was successful")

    def _validate_imputations(self, aggregation_config):
        """Validate the imputation rules in an aggregation config, looping
        through all three types of aggregates. Most of the work here is
        done by _validate_imputation_rule() to check the requirements of
        each imputation rule found
        """
        logger.spam("Validating imputation definitions")
        agg_types = ["aggregates", "categoricals", "array_categoricals"]

        for agg_type in agg_types:
            logger.spam('Validating imputation rules for aggregation type %s', agg_type)
            # base_imp are the top-level rules, `such as aggregates_imputation`
            base_imp = aggregation_config.get(agg_type + "_imputation", {})

            # loop through the individual aggregates
            for agg in aggregation_config.get(agg_type, []):
                logger.spam('Validating imputation rules for aggregation %s', agg)
                # combine any aggregate-level imputation rules with top-level ones
                imp_dict = dict(base_imp, **agg.get("imputation", {}))

                # imputation rules are metric-specific, so check each metric's rule
                for metric in agg["metrics"]:
                    logger.spam('Validating imputation rules for metric: %s', metric)
                    # metric rules may be defined by the metric name (e.g., 'max')
                    # or with the 'all' catch-all, with named metrics taking
                    # precedence. If we fall back to {}, the rule validator will
                    # error out on no metric found.
                    impute_rule = imp_dict.get(metric, imp_dict.get("all", {}))
                    self._validate_imputation_rule(agg_type, impute_rule)
        logger.debug("Validation of imputation definitions was successful")

    def _validate_aggregation(self, aggregation_config):
        logger.spam("Validating aggregation config %s", aggregation_config)
        self._validate_keys(aggregation_config)
        self._validate_aggregates(aggregation_config)
        self._validate_categoricals(aggregation_config.get("categoricals", []))
        self._validate_from_obj(aggregation_config["from_obj"])
        self._validate_time_intervals(aggregation_config["intervals"])
        self._validate_groups(aggregation_config["groups"])
        self._validate_imputations(aggregation_config)
        logger.debug("Validation of aggregation config was successful")

    def _run(self, feature_aggregation_config):
        """Validate a feature aggregation config applied to this object

        The validations range from basic type checks, key presence checks,
        as well as validating the sql in from objects.

        Args:
            feature_aggregation_config (list) all values, except for feature
                date, necessary to instantiate a collate.SpacetimeAggregation

        Raises: ValueError if any part of the config is found to be invalid
        """
        if not feature_aggregation_config:
            raise ValueError(
                dedent(
                    """
            Section: feature_aggregations -
            Section not found. You must define feature aggregations."""
                )
            )
        for aggregation in feature_aggregation_config:
            self._validate_aggregation(aggregation)


class LabelConfigValidator(Validator):
    def _validate_query(self, query):
        if "{as_of_date}" not in query:
            raise ValueError(
                dedent(
                    """
            Section: label_config -
            If 'query' is used as label_config,
            {as_of_date} must be present"""
                )
            )
        if "{label_timespan}" not in query:
            raise ValueError(
                dedent(
                    """
            Section: label_config -
            If 'query' is used as label_config,
            {label_timespan} must be present"""
                )
            )
        bound_query = query.replace("{as_of_date}", "2016-01-01").replace(
            "{label_timespan}", "6month"
        )
        conn = self.db_engine.connect()
        logger.spam("Validating label query via SQL EXPLAIN")
        try:
            conn.execute("explain {}".format(bound_query))
            logger.debug("Validation of label query was successful")
        except Exception as e:
            raise ValueError(
                dedent(
                    """
                Section: label_config -
                given query can not be run with a sample as_of_date and label_timespan.
                query: "{}"
                Full error: {}""".format(
                        query, e
                    )
                )
            )

    @staticmethod
    def _validate_include_missing_labels_in_train_as(missing_label_flag):
        logger.spam("Validating include_missing_labels_in_train")
        if missing_label_flag not in {None, True, False}:
            raise ValueError(
                dedent(
                    """
            Section: label_config -
            The value for 'include_missing_labels_in_train_as', {}, is invalid.
            The key must be either absent, or a boolean value True or False
            Triage only supports binary labels at this time.""".format(
                        missing_label_flag
                    )
                )
            )
        logger.debug("Validation of include_missing_labels_in_train was successful")

    def _run(self, label_config):
        logger.spam("Validating label configuration")
        if not label_config:
            raise ValueError(
                dedent(
                    """
            Section: label_config -
            Section not found. You must define a label config."""
                )
            )

        if "query" not in label_config:
            raise ValueError(
                dedent(
                    """
            Section: label_config -
            key 'query' not found. You must define a label query."""
                )
            )
        if 'name' in label_config and not string_is_tablesafe(label_config['name']):
            raise ValueError("Section: label_config - "
                             "name should only contain lowercase letters, numbers, and underscores")
        self._validate_query(label_config["query"])
        self._validate_include_missing_labels_in_train_as(
            label_config.get("include_missing_labels_in_train_as", None)
        )
        logger.debug("Validation of label configuration was successful")


class CohortConfigValidator(Validator):
    def _run(self, cohort_config):
        logger.spam("Validating of cohort configuration")
        if "query" not in cohort_config:
            raise ValueError(
                dedent(
                    """
            Section: cohort_config -
            key 'query' not found. You must define a cohort query."""
                )
            )
        query = cohort_config["query"]
        if "{as_of_date}" not in query:
            raise ValueError(
                dedent(
                    """
            Section: cohort_config -
            If 'query' is used as cohort_config,
            {as_of_date} must be present"""
                )
            )
        if 'name' in cohort_config and not string_is_tablesafe(cohort_config['name']):
            raise ValueError("Section: cohort_config - "
                             "name should only contain lowercase letters, numbers, and underscores")
        dated_query = query.replace("{as_of_date}", "2016-01-01")
        logger.spam("Validating cohort query via SQL EXPLAIN")
        try:
            self.db_engine.execute(f"explain {dated_query}")
            logger.debug("Validation of cohort query was successful")
        except Exception as e:
            raise ValueError(
                dedent(
                    f"""
                Section: cohort_config -
                given query can not be run with a sample as_of_date .
                query: "{query}"
                Full error: {e}"""
                )
            )
        logger.debug("Validation of cohort configuration was successful")


class FeatureGroupDefinitionValidator(Validator):
    def _validate_prefixes(self, prefix_list):
        """
        Ensure that no prefix starts with another prefix + _ to avoid
        an error with feature group subsets when the group names overlap
        """
        logger.spam("Validating feature group definitions prefixes")
        for prefix1, prefix2 in permutations(prefix_list, 2):
            if prefix2.startswith(prefix1):
                raise ValueError(
                    dedent(
                        """
                Section: feature_group_definition -
                Feature group prefixes must not overlap when using `prefix`: %s and %s"""
                        % (prefix1, prefix2)
                    )
                )
        logger.debug("Validation of feature group definitions prefixes was successful")

    def _run(self, feature_group_definition, feature_aggregation_config):
        logger.spam("Validating of feature group definitions")
        if not isinstance(feature_group_definition, dict):
            raise ValueError(
                dedent(
                    """
            Section: feature_group_definition -
            feature_group_definition must be a dictionary"""
                )
            )

        available_subsetters = (
            architect.feature_group_creator.FeatureGroupCreator.subsetters
        )
        for subsetter_name, value in feature_group_definition.items():
            if subsetter_name not in available_subsetters:
                raise ValueError(
                    dedent(
                        """
                Section: feature_group_definition -
                Unknown feature_group_definition key {} received.
                Available keys are {}""".format(
                            subsetter_name, available_subsetters
                        )
                    )
                )
            if not hasattr(value, "__iter__") or isinstance(value, (str, bytes)):
                raise ValueError(
                    dedent(
                        """
                Section: feature_group_definition -
                feature_group_definition value for {}, {}
                should be a list""".format(
                            subsetter_name, value
                        )
                    )
                )
        logger.debug("Validation of feature group definition was successful")

        if "prefix" in feature_group_definition:
            available_prefixes = {
                aggregation["prefix"] for aggregation in feature_aggregation_config
            }
            bad_prefixes = set(feature_group_definition["prefix"]) - available_prefixes
            if bad_prefixes:
                raise ValueError(
                    dedent(
                        """
                Section: feature_group_definition -
                The following given feature group prefixes: '{}'
                are invalid. Available prefixes from this experiment's feature
                aggregations are: '{}'
                """.format(
                            bad_prefixes, available_prefixes
                        )
                    )
                )
            self._validate_prefixes(feature_group_definition["prefix"])

        if "tables" in feature_group_definition:
            available_tables = {
                aggregation["prefix"] + "_aggregation_imputed"
                for aggregation in feature_aggregation_config
            }
            bad_tables = set(feature_group_definition["tables"]) - available_tables
            if bad_tables:
                raise ValueError(
                    dedent(
                        """
                Section: feature_group_definition -
                The following given feature group tables: '{}'
                are invalid. Available tables from this experiment's feature
                aggregations are: '{}'
                """.format(
                            bad_tables, available_tables
                        )
                    )
                )


class FeatureGroupStrategyValidator(Validator):
    def _run(self, feature_group_strategies):
        logger.spam("Validating feature group strategies")
        if not isinstance(feature_group_strategies, list):
            raise ValueError(
                dedent(
                    """
            Section: feature_group_strategies -
            feature_group_strategies section must be a list"""
                )
            )
        available_strategies = {
            key
            for key in architect.feature_group_mixer.FeatureGroupMixer.strategy_lookup.keys()
        }
        bad_strategies = set(feature_group_strategies) - available_strategies
        if bad_strategies:
            raise ValueError(
                dedent(
                    """
            Section: feature_group_strategies -
            The following given feature group strategies:
            '{}' are invalid. Available strategies are: '{}'
            """.format(
                        bad_strategies, available_strategies
                    )
                )
            )
        logger.debug("Validation of feature group strategies was successful")


class UserMetadataValidator(Validator):
    def _run(self, user_metadata):
        logger.spam("Validating user metadata")
        if not isinstance(user_metadata, dict):
            raise ValueError(
                dedent(
                    """
            Section: user_metadata -
            user_metadata section must be a dict"""
                )
            )
        logger.debug("Validation of user metadata was successful")


class ModelGroupKeysValidator(Validator):
    def _run(self, model_group_keys, user_metadata):
        logger.spam("Validating model group keys")
        if not isinstance(model_group_keys, list):
            raise ValueError(
                dedent(
                    """
            Section: model_group_keys -
            model_group_keys section must be a list"""
                )
            )
        classifier_keys = ["class_path", "parameters"]
        # planner_keys are defined in architect.Planner._make_metadata
        planner_keys = [
            "feature_start_time",
            "end_time",
            "indices",
            "feature_names",
            "feature_groups",
            "label_name",
            "label_type",
            "label_timespan",
            "state",
            "cohort_name",
            "matrix_id",
            "matrix_type",
        ]
        # temporal_keys are defined in
        # timechop.Timechop.generate_matrix_definition
        temporal_keys = [
            "first_as_of_time",
            "last_as_of_time",
            "matrix_info_end_time",
            "as_of_times",
            "training_label_timespan",
            "training_as_of_date_frequency",
            "max_training_history",
        ]
        available_keys = (
            [key for key in user_metadata.keys()]
            + planner_keys
            + temporal_keys
            + classifier_keys
        )
        for model_group_key in model_group_keys:
            if model_group_key not in available_keys:
                raise ValueError(
                    dedent(
                        """
                Section: model_group_keys -
                unknown entry '{}' received. Available keys are {}
                """.format(
                            model_group_key, available_keys
                        )
                    )
                )
        logger.debug("Validation of model group keys was successful")


class GridConfigValidator(Validator):
    def _run(self, grid_config):
        logger.spam("Validating grid configuration")
        if not grid_config:
            raise ValueError(
                dedent(
                    """
            Section: grid_config -
            Section not found. You must define a grid_config."""
                )
            )
        for classpath, parameter_config in grid_config.items():
            if classpath == "sklearn.linear_model.LogisticRegression":
                logger.warning(
                    "sklearn.linear_model.LogisticRegression found in grid. "
                    "This is unscaled and not well-suited for Triage experiments. "
                    "Use triage.component.catwalk.estimators.classifiers.ScaledLogisticRegression "
                    "instead"
                )
            try:
                module_name, class_name = classpath.rsplit(".", 1)
                module = importlib.import_module(module_name)
                cls = getattr(module, class_name)
                for parameters in ParameterGrid(parameter_config):
                    try:
                        cls(**parameters)
                    except Exception as e:
                        raise ValueError(
                            dedent(
                                """
                        Section: grid_config -
                        Unable to instantiate classifier {} with parameters {}, error thrown: {}
                        """.format(
                                    classpath, parameters, e
                                )
                            )
                        )
            except Exception as e:
                raise ValueError(
                    dedent(
                        """
                Section: grid_config -
                Unable to import classifier {}, error thrown: {}
                """.format(
                            classpath, e
                        )
                    )
                )

        logger.debug("Validation of grid configuration was successful")


class PredictionConfigValidator(Validator):
    def _run(self, prediction_config):
        logger.spam("Validating prediction configuration")
        rank_tiebreaker = prediction_config.get("rank_tiebreaker", None)
        # the tiebreaker is optional, so only try and validate if it's there
        if rank_tiebreaker and rank_tiebreaker not in catwalk.utils.AVAILABLE_TIEBREAKERS:
            raise ValueError(
                "Section: prediction - "
                f"given tiebreaker must be in {catwalk.utils.AVAILABLE_TIEBREAKERS}"
            )
        logger.spam("Validation of prediction configuration was successful")


class ScoringConfigValidator(Validator):
    def _run(self, scoring_config):
        logger.spam("Validating scoring configuration")
        if "testing_metric_groups" not in scoring_config:
            logger.warning(
                "Section: scoring - No testing_metric_groups configured. "
                + "Your experiment may run, but you will not have any "
                + "evaluation metrics computed"
            )
        if "training_metric_groups" not in scoring_config:
            logger.warning(
                "Section: scoring - No training_metric_groups configured. "
                + "If training set evaluation metrics are desired, they must be added"
            )
        metric_lookup = catwalk.evaluation.ModelEvaluator.available_metrics
        available_metrics = set(metric_lookup.keys())
        for group in ("testing_metric_groups", "training_metric_groups"):
            for metric_group in scoring_config.get(group, {}):
                given_metrics = set(metric_group["metrics"])
                bad_metrics = given_metrics - available_metrics
                if bad_metrics:
                    raise ValueError(
                        dedent(
                            """Section: scoring -
                        The following given metrics '{}' are unavailable.
                        Available metrics are: '{}'
                        """.format(
                                bad_metrics, available_metrics
                            )
                        )
                    )
                for given_metric in given_metrics:
                    metric_function = metric_lookup[given_metric]
                    if not hasattr(metric_function, "greater_is_better"):
                        raise ValueError(
                            dedent(
                                """Section: scoring -
                        The metric {} does not define the attribute
                        'greater_is_better'. This can only be fixed in the catwalk.metrics
                        module. If you still would like to use this metric, consider
                        submitting a pull request""".format(
                                    given_metric
                                )
                            )
                        )

            if "subsets" in scoring_config:
                for subset in scoring_config["subsets"]:
                    # 1. Validate that all required keys are present
                    if "query" not in subset:
                        raise ValueError(
                            dedent(
                                f"""Section: subsets -
                                The subset {subset} does not have a query key.
                                To run evaluations on a subset, you must
                                include a query that returns a list of distinct
                                entity_ids and has a placeholder for an
                                as_of_date
                                """
                            )
                        )
                    if "name" not in subset:
                        raise ValueError(
                            dedent(
                                f"""Section: subsets -
                                The subset {subset} does not have a name key.
                                Please give a name to your subset. This is used
                                in the namespacing of subset tables created by
                                triage.
                                """
                            )
                        )
                    if not string_is_tablesafe(subset['name']):
                        raise ValueError(
                            dedent(
                                f"""Section: subsets -
                                The subset {subset} name should only contain
                                lowercase letters, numbers, and underscores
                                """
                            )
                        )

                    # 2. Validate that query conforms to the expectations
                    if "{as_of_date}" not in subset["query"]:
                        raise ValueError(
                            dedent(
                                f"""Section: subsets -
                                The subset query {subset["query"]} must
                                include a placeholder for the as_of_date
                                """
                            )
                        )
                    if "entity_id" not in subset["query"]:
                        raise ValueError(
                            dedent(
                                f"""The subset qeury {subset["query"]} must
                                return a list of distinct entity_ids
                                """
                            )
                        )


        logger.debug("Validation of scoring configuration was successful")

class BiasAuditConfigValidator(Validator):
    def _run(self, bias_audit_config):
        logger.spam("Validating bias audit configuration")
        if not bias_audit_config:
            # if empty, that's fine, shortcut out
            return
        if 'from_obj_query' in bias_audit_config and 'from_obj_table' in bias_audit_config:
            raise ValueError(
                dedent(
                    """
                    Section: bias_audit_config -
                    Both 'from_obj_query' and 'from_obj_table' specified .
                    Please only specify one."""
                )
            )
        if 'from_obj_query' not in bias_audit_config and 'from_obj_table' not in bias_audit_config:
            raise ValueError(
                dedent(
                    """
                    Section: bias_audit_config -
                    Neither 'from_obj_query' and 'from_obj_table' specified .
                    Please specify one."""
                )
            )
        for key in [
                "attribute_columns",
                "knowledge_date_column",
                "entity_id_column",
                "ref_groups_method",
        ]:
            if key not in bias_audit_config:
                raise ValueError(
                    dedent(
                        f"""Section: bias_audit_config - '{key} required as key: bias_audit_config config: {bias_audit_config}"""
                    )
                )
        percentile_thresholds = bias_audit_config.get('thresholds', {}).get('percentiles', [])
        if any(threshold < 0 or threshold > 100 for threshold in percentile_thresholds):
            raise ValueError("Section: bias_audit_config - All percentile thresholds must be between 0 and 100")
        logger.debug("Validation of bias audit configuration was successful")

class ExperimentValidator(Validator):
    def run(self, experiment_config):
        logger.spam("Validating experiment configuration")
        TemporalValidator(strict=self.strict).run(
            experiment_config.get("temporal_config", {})
        )
        FeatureAggregationsValidator(self.db_engine, strict=self.strict).run(
            experiment_config.get("feature_aggregations", {})
        )
        LabelConfigValidator(self.db_engine, strict=self.strict).run(
            experiment_config.get("label_config", None)
        )
        CohortConfigValidator(self.db_engine, strict=self.strict).run(
            experiment_config.get("cohort_config", {})
        )
        FeatureGroupDefinitionValidator(strict=self.strict).run(
            experiment_config.get("feature_group_definition", {}),
            experiment_config.get("feature_aggregations", {}),
        )
        FeatureGroupStrategyValidator(strict=self.strict).run(
            experiment_config.get("feature_group_strategies", [])
        )
        UserMetadataValidator(strict=self.strict).run(
            experiment_config.get("user_metadata", {})
        )
        ModelGroupKeysValidator(strict=self.strict).run(
            experiment_config.get("model_group_keys", []),
            experiment_config.get("user_metadata", {}),
        )
        GridConfigValidator(strict=self.strict).run(
            experiment_config.get("grid_config", {})
        )
        PredictionConfigValidator(self.db_engine, strict=self.strict).run(
            experiment_config.get("prediction", {})
        )
        ScoringConfigValidator(strict=self.strict).run(
            experiment_config.get("scoring", {})
        )
        BiasAuditConfigValidator(strict=self.strict).run(
            experiment_config.get("bias_audit_config", {})
        )

        if self.strict:
            logger.success("Experiment validation ran to completion with no errors")
        else:
            logger.warning("Experiment validation complete. All configuration problems have been displayed as warnings")



import os
from datetime import datetime
from tempfile import TemporaryDirectory
import yaml


import testing.postgresql
from triage import create_engine

from triage.component.results_schema import Base
from triage.component.timechop import Timechop
from triage.component.architect.features import (
    FeatureGenerator,
    FeatureDictionaryCreator,
    FeatureGroupCreator,
    FeatureGroupMixer,
)
from triage.component.architect.label_generators import LabelGenerator
from triage.component.architect.entity_date_table_generators import EntityDateTableGenerator
from triage.component.architect.planner import Planner
from triage.component.architect.builders import MatrixBuilder
from triage.component.catwalk.storage import ProjectStorage

from tests.utils import sample_config


def populate_source_data(db_engine):
    cat_complaints = [
        (1, "2010-10-01", 5),
        (1, "2011-10-01", 4),
        (1, "2011-11-01", 4),
        (1, "2011-12-01", 4),
        (1, "2012-02-01", 5),
        (1, "2012-10-01", 4),
        (1, "2013-10-01", 5),
        (2, "2010-10-01", 5),
        (2, "2011-10-01", 5),
        (2, "2011-11-01", 4),
        (2, "2011-12-01", 4),
        (2, "2012-02-01", 6),
        (2, "2012-10-01", 5),
        (2, "2013-10-01", 6),
        (3, "2010-10-01", 5),
        (3, "2011-10-01", 3),
        (3, "2011-11-01", 4),
        (3, "2011-12-01", 4),
        (3, "2012-02-01", 4),
        (3, "2012-10-01", 3),
        (3, "2013-10-01", 4),
    ]

    dog_complaints = [
        (1, "2010-10-01", 5),
        (1, "2011-10-01", 4),
        (1, "2011-11-01", 4),
        (1, "2011-12-01", 4),
        (1, "2012-02-01", 5),
        (1, "2012-10-01", 4),
        (1, "2013-10-01", 5),
        (2, "2010-10-01", 5),
        (2, "2011-10-01", 5),
        (2, "2011-11-01", 4),
        (2, "2011-12-01", 4),
        (2, "2012-02-01", 6),
        (2, "2012-10-01", 5),
        (2, "2013-10-01", 6),
        (3, "2010-10-01", 5),
        (3, "2011-10-01", 3),
        (3, "2011-11-01", 4),
        (3, "2011-12-01", 4),
        (3, "2012-02-01", 4),
        (3, "2012-10-01", 3),
        (3, "2013-10-01", 4),
    ]

    events = [
        (1, 1, "2011-01-01"),
        (1, 1, "2011-06-01"),
        (1, 1, "2011-09-01"),
        (1, 1, "2012-01-01"),
        (1, 1, "2012-01-10"),
        (1, 1, "2012-06-01"),
        (1, 1, "2013-01-01"),
        (1, 0, "2014-01-01"),
        (1, 1, "2015-01-01"),
        (2, 1, "2011-01-01"),
        (2, 1, "2011-06-01"),
        (2, 1, "2011-09-01"),
        (2, 1, "2012-01-01"),
        (2, 1, "2013-01-01"),
        (2, 1, "2014-01-01"),
        (2, 1, "2015-01-01"),
        (3, 0, "2011-01-01"),
        (3, 0, "2011-06-01"),
        (3, 0, "2011-09-01"),
        (3, 0, "2012-01-01"),
        (3, 0, "2013-01-01"),
        (3, 1, "2014-01-01"),
        (3, 0, "2015-01-01"),
    ]

    db_engine.execute(
        """create table cat_complaints (
        entity_id int,
        as_of_date date,
        cat_sightings int
        )"""
    )

    for complaint in cat_complaints:
        db_engine.execute(
            "insert into cat_complaints values (%s, %s, %s)", complaint)

    db_engine.execute(
        """create table dog_complaints (
        entity_id int,
        as_of_date date,
        dog_sightings int
        )"""
    )

    for complaint in dog_complaints:
        db_engine.execute(
            "insert into dog_complaints values (%s, %s, %s)", complaint)

    db_engine.execute(
        """create table events (
        entity_id int,
        outcome int,
        outcome_date date
    )"""
    )

    for event in events:
        db_engine.execute("insert into events values (%s, %s, %s)", event)


def basic_integration_test(
    cohort_names,
    feature_group_create_rules,
    feature_group_mix_rules,
    expected_matrix_multiplier,
    expected_group_lists,
):
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        Base.metadata.create_all(db_engine)
        populate_source_data(db_engine)

        with TemporaryDirectory() as temp_dir:
            chopper = Timechop(
                feature_start_time=datetime(2010, 1, 1),
                feature_end_time=datetime(2014, 1, 1),
                label_start_time=datetime(2011, 1, 1),
                label_end_time=datetime(2014, 1, 1),
                model_update_frequency="1year",
                training_label_timespans=["6months"],
                test_label_timespans=["6months"],
                training_as_of_date_frequencies="1day",
                test_as_of_date_frequencies="3months",
                max_training_histories=["1months"],
                test_durations=["1months"],
            )

            entity_date_table_generator = EntityDateTableGenerator(
                db_engine=db_engine,
                entity_date_table_name="cohort_abcd",
                query="select distinct(entity_id) from events"
            )

            label_generator = LabelGenerator(
                db_engine=db_engine, query=sample_config()[
                    "label_config"]["query"]
            )

            feature_generator = FeatureGenerator(
                db_engine=db_engine, features_schema_name="features", replace=True
            )

            feature_dictionary_creator = FeatureDictionaryCreator(
                db_engine=db_engine, features_schema_name="features"
            )

            feature_group_creator = FeatureGroupCreator(
                feature_group_create_rules)

            feature_group_mixer = FeatureGroupMixer(feature_group_mix_rules)
            project_storage = ProjectStorage(temp_dir)
            planner = Planner(
                feature_start_time=datetime(2010, 1, 1),
                label_names=["outcome"],
                label_types=["binary"],
                cohort_names=cohort_names,
                user_metadata={},
            )

            builder = MatrixBuilder(
                engine=db_engine,
                db_config={
                    "features_schema_name": "features",
                    "labels_schema_name": "public",
                    "labels_table_name": "labels",
                    "cohort_table_name": "cohort_abcd",
                },
                experiment_hash=None,
                matrix_storage_engine=project_storage.matrix_storage_engine(),
                replace=True,
            )

            # chop time
            split_definitions = chopper.chop_time()
            num_split_matrices = sum(
                1 + len(split["test_matrices"]) for split in split_definitions
            )

            # generate as_of_times for feature/label/state generation
            all_as_of_times = []
            for split in split_definitions:
                all_as_of_times.extend(split["train_matrix"]["as_of_times"])
                for test_matrix in split["test_matrices"]:
                    all_as_of_times.extend(test_matrix["as_of_times"])
            all_as_of_times = list(set(all_as_of_times))

            # generate entity_date state table
            entity_date_table_generator.generate_entity_date_table(
                as_of_dates=all_as_of_times)

            # create labels table
            label_generator.generate_all_labels(
                labels_table="labels",
                as_of_dates=all_as_of_times,
                label_timespans=["6months"],
            )

            # create feature table tasks
            # we would use FeatureGenerator#create_all_tables but want to use
            # the tasks dict directly to create a feature dict
            aggregations = feature_generator.aggregations(
                feature_aggregation_config=[
                    {
                        "prefix": "cat",
                        "from_obj": "cat_complaints",
                        "knowledge_date_column": "as_of_date",
                        "aggregates": [
                            {
                                "quantity": "cat_sightings",
                                "metrics": ["count", "avg"],
                                "imputation": {"all": {"type": "mean"}},
                            }
                        ],
                        "intervals": ["1y"],
                        "groups": ["entity_id"],
                    },
                    {
                        "prefix": "dog",
                        "from_obj": "dog_complaints",
                        "knowledge_date_column": "as_of_date",
                        "aggregates_imputation": {
                            "count": {"type": "constant", "value": 7},
                            "sum": {"type": "mean"},
                            "avg": {"type": "zero"},
                        },
                        "aggregates": [
                            {"quantity": "dog_sightings",
                                "metrics": ["count", "avg"]}
                        ],
                        "intervals": ["1y"],
                        "groups": ["entity_id"],
                    },
                ],
                feature_dates=all_as_of_times,
                state_table=entity_date_table_generator.entity_date_table_name,
            )
            feature_table_agg_tasks = feature_generator.generate_all_table_tasks(
                aggregations, task_type="aggregation"
            )

            # create feature aggregation tables
            feature_generator.process_table_tasks(feature_table_agg_tasks)

            feature_table_imp_tasks = feature_generator.generate_all_table_tasks(
                aggregations, task_type="imputation"
            )

            # create feature imputation tables
            feature_generator.process_table_tasks(feature_table_imp_tasks)

            # build feature dictionaries from feature tables and
            # subsetting config
            master_feature_dict = feature_dictionary_creator.feature_dictionary(
                feature_table_names=feature_table_imp_tasks.keys(),
                index_column_lookup=feature_generator.index_column_lookup(
                    aggregations),
            )

            feature_dicts = feature_group_mixer.generate(
                feature_group_creator.subsets(master_feature_dict)
            )

            # figure out what matrices need to be built
            _, matrix_build_tasks = planner.generate_plans(
                split_definitions, feature_dicts
            )

            # go and build the matrices
            builder.build_all_matrices(matrix_build_tasks)

            # super basic assertion: did matrices we expect get created?
            matrices_records = list(
                db_engine.execute(
                    """select matrix_uuid, num_observations, matrix_type
                    from triage_metadata.matrices
                    """
                )
            )
            matrix_directory = os.path.join(temp_dir, "matrices")
            matrices = [path for path in os.listdir(
                matrix_directory) if ".csv" in path]
            metadatas = [
                path for path in os.listdir(matrix_directory) if ".yaml" in path
            ]
            assert len(matrices) == num_split_matrices * \
                expected_matrix_multiplier
            assert len(metadatas) == num_split_matrices * \
                expected_matrix_multiplier
            assert len(matrices) == len(matrices_records)
            feature_group_name_lists = []
            for metadata_path in metadatas:
                with open(os.path.join(matrix_directory, metadata_path)) as f:
                    metadata = yaml.load(f, Loader=yaml.Loader)
                    feature_group_name_lists.append(metadata["feature_groups"])

            for matrix_uuid, num_observations, matrix_type in matrices_records:
                assert matrix_uuid in matrix_build_tasks  # the hashes of the matrices
                assert type(num_observations) is int
                assert matrix_type == matrix_build_tasks[matrix_uuid]["matrix_type"]

            def deep_unique_tuple(l):
                return set([tuple(i) for i in l])

            assert deep_unique_tuple(feature_group_name_lists) == deep_unique_tuple(
                expected_group_lists
            )


def test_integration_simple():
    basic_integration_test(
        cohort_names=["mycohort"],
        feature_group_create_rules={"prefix": ["cat", "dog"]}, # Will be set by defaults.py
        feature_group_mix_rules=["all"],
        # only looking at one state, and one feature group.
        # so we don't multiply timechop's output by anything
        expected_matrix_multiplier=1,
        expected_group_lists=[["prefix: cat", "prefix: dog"]],
    )


def test_integration_feature_grouping():
    basic_integration_test(
        cohort_names=["mycohort"],
        feature_group_create_rules={"prefix": ["cat", "dog"]},
        feature_group_mix_rules=["leave-one-out", "all"],
        # 3 feature groups (cat/dog/cat+dog),
        # so the # of matrices should be each train/test split *3
        expected_matrix_multiplier=3,
        expected_group_lists=[
            ["prefix: cat"],
            ["prefix: cat", "prefix: dog"],
            ["prefix: dog"],
        ],
    )



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from triage.validation_primitives import (
    table_should_exist,
    table_should_have_column,
    column_should_be_timelike
)
import sqlparse


class FromObj:
    def __init__(self, from_obj, name, knowledge_date_column):
        self.from_obj = from_obj
        self.name = name
        self.knowledge_date_column = knowledge_date_column

    @property
    def table(self):
        if self.should_materialize():
            return self.materialized_table
        else:
            return self.from_obj

    @property
    def materialized_table(self):
        return f"{self.name}_from_obj"

    @property
    def create_materialized_table_sql(self):
        return f"create table {self.materialized_table} as (select * from {self.from_obj})"

    @property
    def index_materialized_table_sql(self):
        return f"create index on {self.materialized_table} ({self.knowledge_date_column})"

    @property
    def drop_materialized_table_sql(self):
        return f"drop table if exists {self.materialized_table}"

    def should_materialize(self):
        try:
            (statement,) = sqlparse.parse(self.from_obj)
        except ValueError as exc:
            raise ValueError("Expected exactly one statment to be parsed by sqlparse "
                             f"from from_obj {self.from_obj}.") from exc
        from_obj = statement.token_first(skip_ws=True, skip_cm=True)
        # token_first returns the first 'token' at the top level. This includes any aliases
        # In other words, it's something that you can "select *" from

        # We only want to materialize subqueries. Subqueries need aliases, many other
        # from_objects don't.
        # The first check, 'has_alias', covers this.

        # The real exception is if just a table is specified but has an alias,
        # for easy reference elsewhere.
        # The second check covers this. The 'real name' in these cases is the name
        # of the original table, whereas for a subquery there is no 'real name' besides the alias
        if not isinstance(from_obj, sqlparse.sql.Identifier):
            logger.warning(
                f"Expected {from_obj} to parse as an Identifier. It did not. "
                f"As a result, falling back to *not* materializing raw from object {self.from_obj}"
            )
            return False
        return from_obj.has_alias() and from_obj.get_alias() == from_obj.get_real_name()

    def maybe_materialize(self, db_engine):
        if self.should_materialize():
            logger.spam(f"from_obj in {self.name} looks like a subquery, so creating table")
            db_engine.execute(self.drop_materialized_table_sql)
            db_engine.execute(self.create_materialized_table_sql)
            logger.spam(f"Created table to hold from_obj. New table: {self.materialized_table}")
            self.validate(db_engine)
            db_engine.execute(self.index_materialized_table_sql)
            logger.spam(f"Indexed from_obj table: {self.materialized_table}")
            logger.debug(f"Materialized table {self.materialized_table}")
        else:
            logger.debug(f"from_obj in {self.name} did not look like a subquery, so did not materialize")

    def validate(self, db_engine):
        logger.spam(f"Validating from_obj {self.materialized_table}")
        table_should_exist(self.materialized_table, db_engine)
        logger.spam(f"Table {self.materialized_table} successfully found")
        table_should_have_column(self.materialized_table, 'entity_id', db_engine)
        logger.spam(f"Successfully found entity_id column in {self.materialized_table}")
        table_should_have_column(self.materialized_table, self.knowledge_date_column, db_engine)
        column_should_be_timelike(self.materialized_table, self.knowledge_date_column, db_engine)
        logger.spam(
            f"Successfully found configured knowledge date column in {self.materialized_table}"
        )



import datetime

from triage.component.architect import Planner
from triage.component.architect.feature_group_creator import FeatureGroup


def test_Planner():
    matrix_set_definitions = [
        {
            "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
            "modeling_start_time": datetime.datetime(2010, 1, 1, 0, 0),
            "modeling_end_time": datetime.datetime(2010, 1, 16, 0, 0),
            "train_matrix": {
                "first_as_of_time": datetime.datetime(2010, 1, 1, 0, 0),
                "matrix_info_end_time": datetime.datetime(2010, 1, 6, 0, 0),
                "as_of_times": [
                    datetime.datetime(2010, 1, 1, 0, 0),
                    datetime.datetime(2010, 1, 2, 0, 0),
                    datetime.datetime(2010, 1, 3, 0, 0),
                    datetime.datetime(2010, 1, 4, 0, 0),
                    datetime.datetime(2010, 1, 5, 0, 0),
                ],
            },
            "test_matrices": [
                {
                    "first_as_of_time": datetime.datetime(2010, 1, 6, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 11, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 7, 0, 0),
                        datetime.datetime(2010, 1, 8, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0),
                        datetime.datetime(2010, 1, 10, 0, 0),
                    ],
                }
            ],
        },
        {
            "feature_start_time": datetime.datetime(1990, 1, 1, 0, 0),
            "modeling_start_time": datetime.datetime(2010, 1, 1, 0, 0),
            "modeling_end_time": datetime.datetime(2010, 1, 16, 0, 0),
            "train_matrix": {
                "first_as_of_time": datetime.datetime(2010, 1, 6, 0, 0),
                "matrix_info_end_time": datetime.datetime(2010, 1, 11, 0, 0),
                "as_of_times": [
                    datetime.datetime(2010, 1, 6, 0, 0),
                    datetime.datetime(2010, 1, 7, 0, 0),
                    datetime.datetime(2010, 1, 8, 0, 0),
                    datetime.datetime(2010, 1, 9, 0, 0),
                    datetime.datetime(2010, 1, 10, 0, 0),
                ],
            },
            "test_matrices": [
                {
                    "first_as_of_time": datetime.datetime(2010, 1, 11, 0, 0),
                    "matrix_info_end_time": datetime.datetime(2010, 1, 16, 0, 0),
                    "as_of_times": [
                        datetime.datetime(2010, 1, 11, 0, 0),
                        datetime.datetime(2010, 1, 12, 0, 0),
                        datetime.datetime(2010, 1, 13, 0, 0),
                        datetime.datetime(2010, 1, 14, 0, 0),
                        datetime.datetime(2010, 1, 15, 0, 0),
                    ],
                }
            ],
        },
    ]
    feature_dict_one = FeatureGroup(
        name="first_features",
        features_by_table={"features0": ["f1", "f2"], "features1": ["f1", "f2"]},
    )
    feature_dict_two = FeatureGroup(
        name="second_features",
        features_by_table={"features2": ["f3", "f4"], "features3": ["f5", "f6"]},
    )
    feature_dicts = [feature_dict_one, feature_dict_two]
    planner = Planner(
        feature_start_time=datetime.datetime(2010, 1, 1, 0, 0),
        label_names=["booking"],
        label_types=["binary"],
        cohort_names=["prior_bookings"],
        user_metadata={},
    )

    updated_matrix_definitions, build_tasks = planner.generate_plans(
        matrix_set_definitions, feature_dicts
    )
    # test that it added uuids: we don't much care what they are
    matrix_uuids = []
    for matrix_def in updated_matrix_definitions:
        assert isinstance(matrix_def["train_uuid"], str)
        matrix_uuids.append(matrix_def["train_uuid"])
        for test_uuid in matrix_def["test_uuids"]:
            assert isinstance(test_uuid, str)
    assert len(set(matrix_uuids)) == 4

    # not going to assert anything on the keys (uuids), just get out the values
    build_tasks = build_tasks.values()
    assert len(build_tasks) == 8  # 2 splits * 2 matrices per split * 2 feature dicts

    assert sum(1 for task in build_tasks if task["matrix_type"] == "train") == 4
    assert sum(1 for task in build_tasks if task["matrix_type"] == "test") == 4
    assert (
        sum(1 for task in build_tasks if task["feature_dictionary"] == feature_dict_one)
        == 4
    )
    assert (
        sum(1 for task in build_tasks if task["feature_dictionary"] == feature_dict_two)
        == 4
    )
    assert (
        sum(
            1
            for task in build_tasks
            if task["matrix_metadata"]["feature_groups"] == ["first_features"]
        )
        == 4
    )
    assert (
        sum(
            1
            for task in build_tasks
            if task["matrix_metadata"]["feature_groups"] == ["second_features"]
        )
        == 4
    )
    assert (
        sum(
            1
            for task in build_tasks
            if task["matrix_metadata"]["cohort_name"] == "prior_bookings"
        )
        == 8
    )



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)
import time
from triage.component.catwalk.utils import Batch
from triage.experiments import ExperimentBase

try:
    from rq import Queue
except ImportError:
    logger.error(
        "rq not available. To use RQExperiment, install triage with the RQ extension: "
        "pip install triage[rq]"
    )
    raise


DEFAULT_TIMEOUT = (
    "365d"
)  # We want to basically invalidate RQ's timeouts by setting them each to one year


class RQExperiment(ExperimentBase):
    """An experiment that uses the python-rq library to enqueue tasks and wait for them to finish.

    http://python-rq.org/

    For this experiment to complete, you need some amount of RQ workers running the Triage codebase
    (either on the same machine as the experiment or elsewhere),
    and a Redis instance that both the experiment process and RQ workers can access.

    Args:
        redis_connection (redis.connection): A connection to a Redis instance that
            some rq workers can also access
        sleep_time (int, default 5) How many seconds the process should sleep while
            waiting for RQ results
        queue_kwargs (dict, default {}) Any extra keyword arguments to pass to Queue creation
    """

    def __init__(
        self, redis_connection, sleep_time=5, queue_kwargs=None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.redis_connection = redis_connection
        if queue_kwargs is None:
            queue_kwargs = {}
        self.queue = Queue(connection=self.redis_connection, **queue_kwargs)
        self.sleep_time = sleep_time

    def wait_for(self, jobs):
        """Wait for a list of jobs to complete

        Will run until all jobs are either finished or failed.

        Args:
            jobs (list of rq.Job objects)

        Returns: (list) of job return values
        """
        while True:
            num_done = sum(1 for job in jobs if job.is_finished)
            num_failed = sum(1 for job in jobs if job.is_failed)
            num_pending = sum(
                1 for job in jobs if not job.is_finished and not job.is_failed
            )
            logger.debug(
                f"Report: jobs {num_done} done, {num_failed} failed, {num_pending} pending",
            )
            if num_pending == 0:
                logger.verbose("All jobs completed or failed, returning")
                return [job.result for job in jobs]
            else:
                logger.spam("Sleeping for {self.sleep_time} seconds")
                time.sleep(self.sleep_time)

    def process_query_tasks(self, query_tasks):
        """Run queries by table

        Will run preparation (e.g. create table) and finalize (e.g. create index) tasks
        in the main process,
        but delegate inserts to rq Jobs in batches of 25

        Args: query_tasks (dict) - keys should be table names and values should be dicts.
            Each inner dict should have up to three keys, each with a list of queries:
            'prepare' (setting up the table),
            'inserts' (insert commands to populate the table),
            'finalize' (finishing table setup after all inserts have run)

            Example: {
                'table_one': {
                    'prepare': ['create table table_one (col1 varchar)'],
                    'inserts': [
                        'insert into table_one values (\'a\')',
                        'insert into table_one values (\'b'\')'
                    ]
                    'finalize': ['create index on table_one (col1)']
                }
            }
        """
        for table_name, tasks in query_tasks.items():
            logger.spam(f"Processing features for {table_name}")
            self.feature_generator.run_commands(tasks.get("prepare", []))

            insert_batches = [
                list(task_batch) for task_batch in Batch(tasks.get("inserts", []), 25)
            ]

            jobs = [
                self.queue.enqueue(
                    self.feature_generator.run_commands,
                    insert_batch,
                    job_timeout=DEFAULT_TIMEOUT,
                    result_ttl=DEFAULT_TIMEOUT,
                    ttl=DEFAULT_TIMEOUT,
                )
                for insert_batch in insert_batches
            ]

            self.wait_for(jobs)

            self.feature_generator.run_commands(tasks.get("finalize", []))
            logger.debug(f"{table_name} completed")

    def process_matrix_build_tasks(self, matrix_build_tasks):
        """Run matrix build tasks using RQ

        Args:
            matrix_build_tasks (dict) Keys should be matrix uuids (though not used here),
                values should be dictionaries suitable as kwargs for sending
                to self.matrix_builder.build_matrix

        Returns: (list) of job results for each given task
        """
        jobs = [
            self.queue.enqueue(
                self.matrix_builder.build_matrix,
                job_timeout=DEFAULT_TIMEOUT,
                result_ttl=DEFAULT_TIMEOUT,
                ttl=DEFAULT_TIMEOUT,
                **build_task
            )
            for build_task in matrix_build_tasks.values()
        ]
        return self.wait_for(jobs)

    def process_train_test_batches(self, train_test_batches):
        """Run train tasks using RQ

        Args:
            train_tasks (list) of dictionaries, each representing kwargs suitable
                for self.trainer.process_train_task
        Returns: (list) of job results for each given task
        """
        jobs = [
            self.queue.enqueue(
                self.model_train_tester.process_task,
                job_timeout=DEFAULT_TIMEOUT,
                result_ttl=DEFAULT_TIMEOUT,
                ttl=DEFAULT_TIMEOUT,
                **task
            )
            for batch in train_test_batches
            for task in batch.tasks
        ]
        return self.wait_for(jobs)

    def process_subset_tasks(self, subset_tasks):
        """Run subset tasks using RQ

        Args:
            subset_tasks (list) of dictionaries, each representing kwargs suitable
                for self.subsetter.process_task
        Returns: (list) of job results for each given task
        """
        jobs = [
            self.queue.enqueue(
                self.subsetter.process_task,
                job_timeout=DEFAULT_TIMEOUT,
                result_ttl=DEFAULT_TIMEOUT,
                ttl=DEFAULT_TIMEOUT,
                **task
            )
            for task in subset_tasks
        ]
        return self.wait_for(jobs)



"""empty message

Revision ID: 50e1f1bc2cac
Revises: 0bca1ba9706e
Create Date: 2019-02-19 17:14:31.702012

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '50e1f1bc2cac'
down_revision = '0bca1ba9706e'
branch_labels = None
depends_on = None

def upgrade():
    """
    This upgrade:
        1. adds the model_metadata.subsets table to track evaluation subsets
        2. adds the subset_hash column to the evaluations table, defaulting to
           '' for existing evaluations (on the assumption that they were over
           the whole cohort)
        3. alters (really, drops and re-adds) the primary key for the
           evaluations tables to include the subset_hash
    """
    # 1. Add subsets table
    op.create_table(
        "subsets",
        sa.Column("subset_hash", sa.String(), nullable=False),
        sa.Column("config", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column(
            "created_timestamp",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True
        ),
        sa.PrimaryKeyConstraint("subset_hash"),
        schema="model_metadata",
    )

    # 2. Add subset_hash column
    op.add_column(
        "evaluations",
        sa.Column("subset_hash", sa.String(), nullable=False, server_default=""),
        schema="test_results"
    )
    op.add_column(
        "evaluations",
        sa.Column("subset_hash", sa.String(), nullable=False, server_default=""),
        schema="train_results"
    )

    # 3. Alter primary keys
    # Actual triage databases have been observed with different variats of the
    # primary key name in the train_results schema. To ensure that all
    # databases can be appropriately updated, procedural is used to look up
    # the name of the primary key before droppint it.
    op.drop_constraint("evaluations_pkey", "evaluations", schema="test_results")
    op.execute(
        """
        DO
        $body$
        DECLARE _pkey_name varchar(100) := (
            SELECT conname
              FROM pg_catalog.pg_constraint con
                   INNER JOIN pg_catalog.pg_class rel ON rel.oid = con.conrelid
                   INNER JOIN pg_catalog.pg_namespace nsp ON nsp.oid = con.connamespace
             WHERE rel.relname = 'evaluations'
               AND nspname = 'train_results'
               AND contype = 'p'
        );
        BEGIN
            EXECUTE('ALTER TABLE train_results.evaluations DROP CONSTRAINT ' || _pkey_name);
        END
        $body$
        """
    )

    op.create_primary_key(
        constraint_name="evaluations_pkey",
        table_name="evaluations",
        columns=[
            "model_id",
            "subset_hash",
            "evaluation_start_time",
            "evaluation_end_time",
            "as_of_date_frequency",
            "metric",
            "parameter"
        ],
        schema="test_results",
    )
    op.create_primary_key(
        constraint_name="train_evaluations_pkey",
        table_name="evaluations",
        columns=[
            "model_id",
            "subset_hash",
            "evaluation_start_time",
            "evaluation_end_time",
            "as_of_date_frequency",
            "metric",
            "parameter"
        ],
        schema="train_results",
    )


def downgrade():
    """
    This downgrade revereses the steps of the upgrade:
        1. Alters the primary key on the evaluations tables to exclude
           subset_hash
        2. Drops the subset hash columns from the evaluations tables
        3. Drops the model_metadata.subsets table
    """
    # 1. Alter primary keys
    op.drop_constraint("evaluations_pkey", "evaluations", schema="test_results")
    op.drop_constraint("train_evaluations_pkey", "evaluations", schema="train_results")

    op.create_primary_key(
        name="evaluations_pkey",
        table_name="evaluations",
        columns=[
            "model_id",
            "evaluation_start_time",
            "evaluation_end_time",
            "as_of_date_frequency",
            "metric",
            "parameter"
        ],
        schema="test_results",
    )
    op.create_primary_key(
        name="train_evaluations_pkey",
        table_name="evaluations",
        columns=[
            "model_id",
            "evaluation_start_time",
            "evaluation_end_time",
            "as_of_date_frequency",
            "metric",
            "parameter"
        ],
        schema="train_results",
    )

    # 2. Drop subset_hash columns
    op.drop_column("evaluations", "subset_hash", schema="train_results")
    op.drop_column("evaluations", "subset_hash", schema="test_results")

    # 3. Drop subsets table
    op.drop_table("subsets", schema="model_metadata")



from unittest.mock import patch

import numpy as np
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.audition.regrets import SelectionRulePicker, SelectionRulePlotter
from triage.component.audition.selection_rules import (
    best_current_value,
    best_average_value,
    BoundSelectionRule,
)

from .utils import create_sample_distance_table


def test_selection_rule_picker():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        distance_table, model_groups = create_sample_distance_table(engine)

        def pick_spiky(df, train_end_time):
            return [model_groups["spiky"].model_group_id]

        selection_rule_picker = SelectionRulePicker(
            distance_from_best_table=distance_table
        )

        results = selection_rule_picker.results_for_rule(
            bound_selection_rule=BoundSelectionRule(
                descriptive_name="spiky", function=pick_spiky, args={}
            ),
            model_group_ids=[mg.model_group_id for mg in model_groups.values()],
            train_end_times=["2014-01-01", "2015-01-01", "2016-01-01"],
            regret_metric="precision@",
            regret_parameter="100_abs",
        )
        assert [result["dist_from_best_case_next_time"] for result in results] == [
            0.19,
            0.3,
            0.12,
        ]
        assert [result["raw_value"] for result in results] == [0.45, 0.84, 0.45]


def test_selection_rule_picker_with_args():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        distance_table, model_groups = create_sample_distance_table(engine)

        def pick_highest_avg(df, train_end_time, metric, parameter):
            assert len(df["train_end_time"].unique()) == 2
            subsetted = df[(df["metric"] == metric) & (df["parameter"] == parameter)]
            mean = subsetted.groupby(["model_group_id"])["raw_value"].mean()
            return [mean.nlargest(1).index[0]]

        selection_rule_picker = SelectionRulePicker(
            distance_from_best_table=distance_table
        )
        regrets = [
            result["dist_from_best_case_next_time"]
            for result in selection_rule_picker.results_for_rule(
                bound_selection_rule=BoundSelectionRule(
                    descriptive_name="pick_highest_avg",
                    function=pick_highest_avg,
                    args={"metric": "recall@", "parameter": "100_abs"},
                ),
                model_group_ids=[mg.model_group_id for mg in model_groups.values()],
                train_end_times=["2015-01-01"],
                regret_metric="precision@",
                regret_parameter="100_abs",
            )
        ]
        # picking the highest avg recall will pick 'spiky' for this time
        assert regrets == [0.3]


def test_SelectionPlotter_create_plot_dataframe():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        distance_table, model_groups = create_sample_distance_table(engine)
        plotter = SelectionRulePlotter(
            selection_rule_picker=SelectionRulePicker(distance_table)
        )
        plot_df = plotter.create_plot_dataframe(
            bound_selection_rules=[
                BoundSelectionRule(
                    descriptive_name="best_current_precision",
                    function=best_current_value,
                    args={"metric": "precision@", "parameter": "100_abs"},
                ),
                BoundSelectionRule(
                    descriptive_name="best_avg_precision",
                    function=best_average_value,
                    args={"metric": "precision@", "parameter": "100_abs"},
                ),
            ],
            model_group_ids=[mg.model_group_id for mg in model_groups.values()],
            train_end_times=["2014-01-01", "2015-01-01", "2016-01-01"],
            regret_metric="precision@",
            regret_parameter="100_abs",
        )
        # assert that we have the right # of columns and a row for each % diff value
        assert plot_df.shape == (100 * 2, 3)

        # both selection rules have a regret lower than 70
        for value in plot_df[plot_df["regret"] == 0.70]["pct_of_time"].values:
            assert np.isclose(value, 1.0)

        # best avg precision rule should be within 0.14 1/3 of the time
        for value in plot_df[
            (plot_df["regret"] == 0.14)
            & (plot_df["selection_rule"] == "best_avg_precision")
        ]["pct_of_time"].values:
            assert np.isclose(value, 1.0 / 3)


def test_SelectionPlotter_plot():
    with patch("triage.component.audition.regrets.plot_cats") as plot_patch:
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            distance_table, model_groups = create_sample_distance_table(engine)
            plotter = SelectionRulePlotter(
                selection_rule_picker=SelectionRulePicker(distance_table)
            )
            plotter.plot_all_selection_rules(
                bound_selection_rules=[
                    BoundSelectionRule(
                        descriptive_name="best_current_precision",
                        function=best_current_value,
                        args={"metric": "precision@", "parameter": "100_abs"},
                    ),
                    BoundSelectionRule(
                        descriptive_name="best_avg_precision",
                        function=best_average_value,
                        args={"metric": "precision@", "parameter": "100_abs"},
                    ),
                ],
                model_group_ids=[mg.model_group_id for mg in model_groups.values()],
                train_end_times=["2014-01-01", "2015-01-01", "2016-01-01"],
                regret_metric="precision@",
                regret_parameter="100_abs",
            )
        assert plot_patch.called
        args, kwargs = plot_patch.call_args
        assert "regret" in kwargs["frame"]
        assert "pct_of_time" in kwargs["frame"]
        assert kwargs["x_col"] == "regret"
        assert kwargs["y_col"] == "pct_of_time"



from triage.component.catwalk.individual_importance import (
    IndividualImportanceCalculator,
    IndividualImportanceCalculatorNoOp,
)
from tests.utils import (
    rig_engines,
    fake_trained_model,
    matrix_creator,
    matrix_metadata_creator,
    get_matrix_store,
)

from unittest.mock import patch


def sample_individual_importance_strategy(
    db_engine, model_id, as_of_date, test_matrix_store, n_ranks
):
    return [
        {
            "entity_id": 1,
            "feature_value": 0.5,
            "feature_name": "m_feature",
            "score": 0.5,
        },
        {
            "entity_id": 1,
            "feature_value": 0.5,
            "feature_name": "k_feature",
            "score": 0.5,
        },
    ]


@patch.dict(
    "triage.component.catwalk.individual_importance.CALCULATE_STRATEGIES",
    {"sample": sample_individual_importance_strategy},
)
def test_calculate_and_save():
    with rig_engines() as (db_engine, project_storage):
        train_store = get_matrix_store(
            project_storage,
            matrix_creator(),
            matrix_metadata_creator(matrix_type="train"),
        )
        test_store = get_matrix_store(
            project_storage,
            matrix_creator(),
            matrix_metadata_creator(matrix_type="test"),
        )
        calculator = IndividualImportanceCalculator(
            db_engine, methods=["sample"], replace=False
        )
        # given a trained model
        # and a test matrix
        _, model_id = fake_trained_model(db_engine, train_matrix_uuid=train_store.uuid)
        # i expect to be able to call calculate and save
        calculator.calculate_and_save_all_methods_and_dates(model_id, test_store)
        # and find individual importances in the results schema afterwards
        records = [
            row
            for row in db_engine.execute(
                """select entity_id, as_of_date
            from test_results.individual_importances
            join triage_metadata.models using (model_id)"""
            )
        ]
        assert len(records) > 0
        # and that when run again, has the same result
        calculator.calculate_and_save_all_methods_and_dates(model_id, test_store)
        new_records = [
            row
            for row in db_engine.execute(
                """select entity_id, as_of_date
            from test_results.individual_importances
            join triage_metadata.models using (model_id)"""
            )
        ]
        assert len(records) == len(new_records)
        assert records == new_records


@patch.dict(
    "triage.component.catwalk.individual_importance.CALCULATE_STRATEGIES",
    {"sample": sample_individual_importance_strategy},
)
def test_calculate_and_save_noop():
    with rig_engines() as (db_engine, project_storage):
        train_store = get_matrix_store(
            project_storage,
            matrix_creator(),
            matrix_metadata_creator(matrix_type="train"),
        )
        test_store = get_matrix_store(
            project_storage,
            matrix_creator(),
            matrix_metadata_creator(matrix_type="test"),
        )
        calculator = IndividualImportanceCalculatorNoOp()
        # given a trained model
        # and a test matrix
        _, model_id = fake_trained_model(db_engine, train_matrix_uuid=train_store.uuid)
        # i expect to be able to call calculate and save
        calculator.calculate_and_save_all_methods_and_dates(model_id, test_store)
        # and find individual importances in the results schema afterwards
        records = [
            row
            for row in db_engine.execute(
                """select entity_id, as_of_date
            from test_results.individual_importances
            join triage_metadata.models using (model_id)"""
            )
        ]
        assert len(records) == 0
        # and that when run again, has the same result
        calculator.calculate_and_save_all_methods_and_dates(model_id, test_store)
        new_records = [
            row
            for row in db_engine.execute(
                """select entity_id, as_of_date
            from test_results.individual_importances
            join triage_metadata.models using (model_id)"""
            )
        ]
        assert len(records) == len(new_records)
        assert records == new_records



from datetime import datetime

import factory
import pandas as pd
import testing.postgresql
from sqlalchemy import create_engine

from tests.results_tests.factories import (
    EvaluationFactory,
    ModelFactory,
    ModelGroupFactory,
    init_engine,
    session,
)

from triage.component.audition.pre_audition import PreAudition
from triage.component.catwalk.db import ensure_db


def test_PreAudition():
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        init_engine(db_engine)
        # set up data, randomly generated by the factories but conforming
        # generally to what we expect triage_metadata schema data to look like
        num_model_groups = 10
        model_types = [
            "classifier type {}".format(i) for i in range(0, num_model_groups)
        ]
        model_configs = [
            {"label_definition": "label_1"}
            if i % 2 == 0
            else {"label_definition": "label_2"}
            for i in range(0, num_model_groups)
        ]
        model_groups = [
            ModelGroupFactory(model_type=model_type, model_config=model_config)
            for model_type, model_config in zip(model_types, model_configs)
        ]
        train_end_times = [
            datetime(2013, 1, 1),
            datetime(2013, 7, 1),
            datetime(2014, 1, 1),
            datetime(2014, 7, 1),
            datetime(2015, 1, 1),
            datetime(2015, 7, 1),
            datetime(2016, 7, 1),
            datetime(2016, 1, 1),
        ]

        models = [
            ModelFactory(model_group_rel=model_group, train_end_time=train_end_time)
            for model_group in model_groups
            for train_end_time in train_end_times
        ]
        metrics = [
            ("precision@", "100_abs"),
            ("recall@", "100_abs"),
            ("precision@", "50_abs"),
            ("recall@", "50_abs"),
            ("fpr@", "10_pct"),
        ]

        class ImmediateEvalFactory(EvaluationFactory):
            evaluation_start_time = factory.LazyAttribute(
                lambda o: o.model_rel.train_end_time
            )

        for model in models:
            for (metric, parameter) in metrics:
                ImmediateEvalFactory(
                    model_rel=model, metric=metric, parameter=parameter
                )

        session.commit()

        pre_aud = PreAudition(db_engine)

        # Expect the number of model groups with label_1
        assert len(pre_aud.get_model_groups_from_label("label_1")['model_groups']) == sum(
            [x["label_definition"] == "label_1" for x in model_configs]
        )

        # Expect no baseline model groups
        assert len(pre_aud.get_model_groups_from_label("label_1")['baseline_model_groups']) == 0

        # Expect the number of model groups with certain experiment_hash
        experiment_hash = list(
            pd.read_sql(
                """SELECT experiment_hash
                FROM triage_metadata.models
                JOIN triage_metadata.experiment_models using (model_hash)
                limit 1""",
                con=db_engine,
            )["experiment_hash"]
        )[0]
        assert len(pre_aud.get_model_groups_from_experiment(experiment_hash)['model_groups']) == 1

        # Expect the number of model groups for customs SQL
        query = """
            SELECT DISTINCT(model_group_id)
            FROM triage_metadata.models
            JOIN triage_metadata.experiment_models using (model_hash)
            WHERE train_end_time >= '2013-01-01'
            AND experiment_hash = '{}'
        """.format(
            experiment_hash
        )
        assert len(pre_aud.get_model_groups(query)) == 1
        # Expect the number of train_end_times after 2014-01-01
        assert len(pre_aud.get_train_end_times(after="2014-01-01")) == 6

        query = """
            SELECT DISTINCT train_end_time
            FROM triage_metadata.models
            WHERE model_group_id IN ({})
                AND train_end_time >= '2014-01-01'
            ORDER BY train_end_time
            """.format(
            ", ".join(map(str, pre_aud.model_groups))
        )

        assert len(pre_aud.get_train_end_times(query=query)) == 6



"""Metric definitions

Mostly just wrappers around sklearn.metrics functions, these functions
implement a generalized interface to metric calculations that can be stored
as a scalar in the database.

All functions should take four parameters:
predictions_proba (1d array-like) Prediction probabilities
predictions_binary (1d array-like) Binarized predictions
labels (1d array-like) Ground truth target values
parameters (dict) Any needed hyperparameters in the implementation

All functions should be wrapped with @Metric to define the optimal direction

All functions should return: (float) the resulting score

Functions defined here are meant to be used in ModelEvaluator.available_metrics

"""
from sklearn import metrics
from sklearn.metrics import confusion_matrix
import numpy as np


class Metric:
    """decorator for metrics: result will be a callable metric with an
    `greater_is_better` parameter defined as either True or False
    depending on whether larger or smaller metric values indicate
    better models.
    """

    def __init__(self, greater_is_better):
        if greater_is_better not in (True, False):
            raise ValueError("greater_is_better must be True or False")
        self.greater_is_better = greater_is_better

    def __call__(self, function, *params, **kwparams):
        class DecoratedMetric:
            def __init__(self, greater_is_better, function):
                self.greater_is_better = greater_is_better
                self.function = function
                self.__name__ = function.__name__
                self.__doc__ = function.__doc__

            def __call__(self, *params, **kwparams):
                return self.function(*params, **kwparams)

        return DecoratedMetric(self.greater_is_better, function)


@Metric(greater_is_better=True)
def precision(_, predictions_binary, labels, parameters):
    return metrics.precision_score(labels, predictions_binary, **parameters)


@Metric(greater_is_better=True)
def recall(_, predictions_binary, labels, parameters):
    return metrics.recall_score(labels, predictions_binary, **parameters)


@Metric(greater_is_better=True)
def fbeta(_, predictions_binary, labels, parameters):
    return metrics.fbeta_score(labels, predictions_binary, **parameters)


@Metric(greater_is_better=True)
def f1(_, predictions_binary, labels, parameters):
    return metrics.f1_score(labels, predictions_binary, **parameters)


@Metric(greater_is_better=True)
def accuracy(_, predictions_binary, labels, parameters):
    return metrics.accuracy_score(labels, predictions_binary, **parameters)


@Metric(greater_is_better=True)
def roc_auc(predictions_proba, _, labels, parameters):
    return metrics.roc_auc_score(labels, predictions_proba)


@Metric(greater_is_better=True)
def avg_precision(predictions_proba, _, labels, parameters):
    return metrics.average_precision_score(labels, predictions_proba)


@Metric(greater_is_better=True)
def true_positives(_, predictions_binary, labels, parameters):
    # If all labels false
    if not any(labels):
        return 0
    # If all labels true and all predictions 1
    elif all(labels) and all(i == 1 for i in predictions_binary):
        return 1
    else:
        return int(confusion_matrix(labels, predictions_binary)[1, 1])


@Metric(greater_is_better=False)
def false_positives(_, predictions_binary, labels, parameters):
    # If all labels false
    if not any(labels):
        return 0
    # If all labels true and all predictions 1
    elif all(labels) and all(i == 1 for i in predictions_binary):
        return 0
    else:
        return int(confusion_matrix(labels, predictions_binary)[0, 1])


@Metric(greater_is_better=True)
def true_negatives(_, predictions_binary, labels, parameters):
    # If all labels false and all predictions 0
    if not any(labels) and all(i == 0 for i in predictions_binary):
        return 1
    # If all labels true
    elif all(labels):
        return 0
    else:
        return int(confusion_matrix(labels, predictions_binary)[0, 0])


@Metric(greater_is_better=False)
def false_negatives(_, predictions_binary, labels, parameters):
    # If all labels false and all predictions 0
    if not any(labels) and all(i == 0 for i in predictions_binary):
        return 0
    # If all labels true
    elif all(labels):
        return 0
    else:
        return int(confusion_matrix(labels, predictions_binary)[1, 0])


@Metric(greater_is_better=False)
def fpr(_, predictions_binary, labels, parameters):
    fp = false_positives(_, predictions_binary, labels, parameters)
    return float(fp / (len(labels) - np.count_nonzero(labels)))


class UnknownMetricError(ValueError):
    """Signifies that a metric name was passed, but no matching computation
    function is available
    """



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from triage.component.catwalk.utils import save_db_objects
from triage.component.results_schema import IndividualImportance

from .uniform import uniform_distribution


CALCULATE_STRATEGIES = {"uniform": uniform_distribution}


class IndividualImportanceCalculatorNoOp:
    def calculate_and_save_all_methods_and_dates(self, model_id, test_matrix_store):
        logger.notice(
            "No individual feature importance configuration is available, so no individual feature importance will be created"
        )

    def calculate_and_save(self, model_id, test_matrix_store, method, as_of_date):
        logger.notice(
            "No individual feature importance configuration is available, so no individual feature importance will be created"
        )


    def save(self, importance_records, model_id, as_of_date, method_name):
        logger.notice(
            "No individual feature importance configuration is available, so no individual feature importance will be created"
        )



class IndividualImportanceCalculator:
    """Calculates and saves individual importance scores and rankings using different methods

    Args:
        db_engine (sqlalchemy.engine)
        n_ranks (int) Number of ranks to calculate and save. Defaults to 5
        methods (list) Strings corresponding to individual importance methods  ,
            present in CALCULATE_STRATEGIES that should be called.
            Defaults to ['uniform']
        replace (bool) Whether to replace old records or reuse them.
    """

    def __init__(self, db_engine, n_ranks=5, methods=["uniform"], replace=True):
        self.db_engine = db_engine
        self.n_ranks = n_ranks
        self.methods = methods
        self.replace = replace

    def _num_existing_importances(self, model_id, as_of_date, method):
        return [
            row[0]
            for row in self.db_engine.execute(
                """select count(*) from test_results.individual_importances
            where model_id = %s
            and as_of_date = %s
            and method = %s""",
                model_id,
                as_of_date,
                method,
            )
        ][0]

    def _needs_new_importances(self, model_id, as_of_date, method, matrix_store):
        """Determines whether or not importances matching the arguments are present in the database

        To do this, we check how many importances exist for the given arguments
        and compare with the number of importances that the given matrix will produce.

        Args:
            model_id (int) A model id, expected to be present in test_results.models
            as_of_date (datetime or string) The date to produce individual importances as of
            method (string) The name of a method to use to produce individual importances
            matrix_store (catwalk.storage.MatrixStore) The test matrix

        Returns: (bool) whether or not there are fewer importances in the db than the matrix
        """
        existing_importances = self._num_existing_importances(
            model_id, as_of_date, method
        )
        expected_importances = matrix_store.num_entities * self.n_ranks
        logger.debug(
            "model_id=%s/as_of_date=%s/method=%s: found %s importances",
            model_id,
            as_of_date,
            method,
            existing_importances,
        )
        logger.debug(
            "matrix_uuid=%s/n_ranks=%s: expect %s importances",
            matrix_store.uuid,
            self.n_ranks,
            expected_importances,
        )
        return existing_importances < expected_importances

    def calculate_and_save_all_methods_and_dates(self, model_id, test_matrix_store):
        """Calculate and save individual importances for the given model and test matrix

        Args:
            model_id (int) A model id, expected to be present in test_results.models
            test_matrix_store (catwalk.storage.MatrixStore) The test matrix
        """
        for method in self.methods:
            for as_of_date in test_matrix_store.as_of_dates:
                self.calculate_and_save(model_id, test_matrix_store, method, as_of_date)

    def calculate_and_save(self, model_id, test_matrix_store, method, as_of_date):
        """Calculate and save importances for a given model, test matrix, method, and date

        Args:
            model_id (int) A model id, expected to be present in test_results.models
            test_matrix_store (catwalk.storage.MatrixStore) The test matrix
            method_name (string) The name of a method to use to produce individual importances
                Expected to be present in CALCULATE_STRATEGIES
            as_of_date (datetime or string) The date to produce individual importances as of
        """
        if not self.replace and not self._needs_new_importances(
            model_id, as_of_date, method, test_matrix_store
        ):
            logger.debug(
                "Found as many or more individual importances "
                + "for model_id=%s/as_of_date=%s/method=%s, skipping",
                model_id,
                as_of_date,
                method,
            )
        else:
            importance_records = CALCULATE_STRATEGIES[method](
                self.db_engine, model_id, as_of_date, test_matrix_store, self.n_ranks
            )
            self.save(
                importance_records=importance_records,
                model_id=model_id,
                as_of_date=as_of_date,
                method_name=method,
            )

    def save(self, importance_records, model_id, as_of_date, method_name):
        """Saves computed individual feature importance records to the database.
        Will delete any records beforehand matching the model_id, as_of_date, and method_name

        Args:
            importance_records (list) Individual importances.
                Each entry should be a dict with keys 'entity_id' 'feature_name'
                and 'feature_value' corresponding to the correct values for the
                model_id, as_of_date, and method
            model_id (int) A model id, expected to be present in test_results.models
            as_of_date (datetime or string) The as_of_date matching the importance records
            method_name (string) The name of the method that produced the importance records

        """
        self.db_engine.execute(
            """delete from test_results.individual_importances
            where model_id = %s
            and as_of_date = %s
            and method = %s""",
            model_id,
            as_of_date,
            method_name,
        )
        record_stream = (
            IndividualImportance(
                model_id=int(model_id),
                entity_id=int(importance_record["entity_id"]),
                as_of_date=as_of_date,
                feature=importance_record["feature_name"],
                feature_value=importance_record["feature_value"],
                method=method_name,
                importance_score=float(importance_record["score"]),
            )
            for importance_record in importance_records
        )
        save_db_objects(self.db_engine, record_stream)



import pandas as pd
import os
from .plotting import plot_cats


class SelectionRulePerformancePlotter:
    """Plot regrets over time

    Generates a simple line chart with audition.BoundSelectionRule objects as
    lines, time periods to be analyzed as the x-axis, and the regret of that
    selection rule as the y-axis.

    Args:
        selection_rule_picker (audition.SelectionRulePicker) An object that can
            simulate the effects of picking a selection rule
    """

    def __init__(self, selection_rule_picker, directory=None):
        self.selection_rule_picker = selection_rule_picker
        self.results_for_rule = None
        self.directory = directory

    def plot(
        self,
        bound_selection_rules,
        regret_metric,
        regret_parameter,
        model_group_ids,
        train_end_times,
        plot_type="regret",
    ):
        """Generate a selection rule performance plot for one metric

        Args:
            bound_selection_rules (list of .selection_rules.BoundSelectionRule)
               Selection rule/parameter combinations to plot
            regret_metric (string) -- model evaluation metric, such as 'precision@'
            regret_parameter (string) -- model evaluation metric parameter,
                    such as '300_abs'
            model_group_ids (list of integers) The model group ids to include
                in calculating the data
            train_end_times (list of timestamps) The timestamps to include in
                calculating the data
            plot_type (string) The plot type to show (either 'regret' or 'metric')
        """
        df = self.generate_plot_data(
            bound_selection_rules=bound_selection_rules,
            model_group_ids=model_group_ids,
            train_end_times=train_end_times,
            regret_metric=regret_metric,
            regret_parameter=regret_parameter,
        )
        if plot_type == "regret":
            self.regret_plot_from_dataframe(
                metric=regret_metric, parameter=regret_parameter, df=df
            )
        elif plot_type == "metric":
            self.raw_next_time_plot_from_dataframe(
                metric=regret_metric, parameter=regret_parameter, df=df
            )
        else:
            raise ValueError("Plot type must be either regret or metric")

    def generate_plot_data(
        self,
        bound_selection_rules,
        model_group_ids,
        train_end_times,
        regret_metric,
        regret_parameter,
    ):
        """Create a dataframe suitable for plotting regrets over time

        Args:
            bound_selection_rules (list of .selection_rules.BoundSelectionRule)
               Selection rule/parameter combinations to plot
            model_group_ids (list of integers) The model group ids to include
                in calculating the data
            train_end_times (list of timestamps) The timestamps to include in
                calculating the data
            regret_metric (string) -- model evaluation metric, such as 'precision@'
            regret_parameter (string) -- model evaluation metric parameter,
                such as '300_abs'

        Returns: (pandas.DataFrame) A dataframe with columns 'regret',
            'train_end_time', and 'selection_rule'
        """
        accumulator = list()
        for selection_rule in bound_selection_rules:
            results = self.selection_rule_picker.results_for_rule(
                selection_rule,
                model_group_ids,
                train_end_times,
                regret_metric,
                regret_parameter,
            )
            for result in results:
                accumulator.append(
                    {
                        "train_end_time": result["train_end_time"],
                        "regret": result["dist_from_best_case_next_time"],
                        "selection_rule": selection_rule.descriptive_name,
                        "raw_value_next_time": result["raw_value_next_time"],
                        "model_group_id": result["model_group_id"],
                    }
                )
        return pd.DataFrame.from_records(accumulator)

    def regret_plot_from_dataframe(self, metric, parameter, df, **plt_format_args):
        """Generate a regret-over-time plot from a given dataframe

        Args:
            metric (string) -- model evaluation metric, such as 'precision@'
            parameter (string) -- model evaluation metric parameter,
                such as '300_abs'
            df (pandas.DataFrame) -- A dataframe preresenting regrets over time.
                 Should have columns 'regret', 'train_end_time', 'selection_rule'
        """
        cat_col = "selection_rule"
        plt_title = "Regret for {} {} over time".format(metric, parameter)
        if self.directory:
            path_to_save = os.path.join(
                self.directory, f"regret_over_time_{metric}{parameter}.png"
            )
        else:
            path_to_save = None
        plot_cats(
            frame=df,
            x_col="train_end_time",
            y_col="regret",
            cat_col=cat_col,
            grp_col="selection_rule",
            title=plt_title,
            x_label="train end time",
            y_label="regret in {}".format(metric),
            x_lim=(df["train_end_time"].min(), df["train_end_time"].max()),
            path_to_save=path_to_save,
            alpha=1.0,
            **plt_format_args,
        )

    def raw_next_time_plot_from_dataframe(
        self, metric, parameter, df, **plt_format_args
    ):
        """Generate a 'raw-value-next-time' plot from a given dataframe

        Args:
            metric (string) -- model evaluation metric, such as 'precision@'
            parameter (string) -- model evaluation metric parameter,
                such as '300_abs'
            df (pandas.DataFrame) -- A dataframe that contains raw metric values
                over time given selection rules.
                Should have columns 'raw_value_next_time', 'train_end_time', 'selection_rule'
        """
        cat_col = "selection_rule"
        plt_title = "{} {} next time".format(metric, parameter)
        if self.directory:
            path_to_save = os.path.join(
                self.directory, f"{metric}{parameter}_next_time.png"
            )
        else:
            path_to_save = None
        plot_cats(
            frame=df,
            x_col="train_end_time",
            y_col="raw_value_next_time",
            cat_col=cat_col,
            grp_col="selection_rule",
            title=plt_title,
            x_label="train end time",
            y_label="value of {} next time".format(metric),
            x_lim=(df["train_end_time"].min(), df["train_end_time"].max()),
            path_to_save=path_to_save,
            alpha=1.0,
            **plt_format_args,
        )



from triage.component.audition.distance_from_best import DistanceFromBestTable
from triage.component.catwalk.db import ensure_db

from tests.results_tests.factories import (
    ModelFactory,
    ModelGroupFactory,
    init_engine,
    session,
)


def create_sample_distance_table(engine):
    ensure_db(engine)
    init_engine(engine)
    model_groups = {
        "stable": ModelGroupFactory(model_type="myStableClassifier"),
        "spiky": ModelGroupFactory(model_type="mySpikeClassifier"),
    }

    class StableModelFactory(ModelFactory):
        model_group_rel = model_groups["stable"]

    class SpikyModelFactory(ModelFactory):
        model_group_rel = model_groups["spiky"]

    models = {
        "stable_3y_ago": StableModelFactory(train_end_time="2014-01-01"),
        "stable_2y_ago": StableModelFactory(train_end_time="2015-01-01"),
        "stable_1y_ago": StableModelFactory(train_end_time="2016-01-01"),
        "spiky_3y_ago": SpikyModelFactory(train_end_time="2014-01-01"),
        "spiky_2y_ago": SpikyModelFactory(train_end_time="2015-01-01"),
        "spiky_1y_ago": SpikyModelFactory(train_end_time="2016-01-01"),
    }
    session.commit()
    distance_table = DistanceFromBestTable(
        db_engine=engine, models_table="models", distance_table="dist_table", agg_type="worst"
    )
    distance_table._create()
    stable_grp = model_groups["stable"].model_group_id
    spiky_grp = model_groups["spiky"].model_group_id
    stable_3y_id = models["stable_3y_ago"].model_id
    stable_3y_end = models["stable_3y_ago"].train_end_time
    stable_2y_id = models["stable_2y_ago"].model_id
    stable_2y_end = models["stable_2y_ago"].train_end_time
    stable_1y_id = models["stable_1y_ago"].model_id
    stable_1y_end = models["stable_1y_ago"].train_end_time
    spiky_3y_id = models["spiky_3y_ago"].model_id
    spiky_3y_end = models["spiky_3y_ago"].train_end_time
    spiky_2y_id = models["spiky_2y_ago"].model_id
    spiky_2y_end = models["spiky_2y_ago"].train_end_time
    spiky_1y_id = models["spiky_1y_ago"].model_id
    spiky_1y_end = models["spiky_1y_ago"].train_end_time
    distance_rows = [
        (
            stable_grp,
            stable_3y_end,
            "precision@",
            "100_abs",
            0.5,
            0.6,
            0.1,
            0.5,
            0.15,
        ),
        (
            stable_grp,
            stable_2y_end,
            "precision@",
            "100_abs",
            0.5,
            0.84,
            0.34,
            0.5,
            0.18,
        ),
        (
            stable_grp,
            stable_1y_end,
            "precision@",
            "100_abs",
            0.46,
            0.67,
            0.21,
            0.5,
            0.11,
        ),
        (
            spiky_grp,
            spiky_3y_end,
            "precision@",
            "100_abs",
            0.45,
            0.6,
            0.15,
            0.5,
            0.19,
        ),
        (
            spiky_grp,
            spiky_2y_end,
            "precision@",
            "100_abs",
            0.84,
            0.84,
            0.0,
            0.5,
            0.3,
        ),
        (
            spiky_grp,
            spiky_1y_end,
            "precision@",
            "100_abs",
            0.45,
            0.67,
            0.22,
            0.5,
            0.12,
        ),
        (
            stable_grp,
            stable_3y_end,
            "recall@",
            "100_abs",
            0.4,
            0.4,
            0.0,
            0.4,
            0.0,
        ),
        (
            stable_grp,
            stable_2y_end,
            "recall@",
            "100_abs",
            0.5,
            0.5,
            0.0,
            0.5,
            0.0,
        ),
        (
            stable_grp,
            stable_1y_end,
            "recall@",
            "100_abs",
            0.6,
            0.6,
            0.0,
            0.6,
            0.0,
        ),
        (
            spiky_grp,
            spiky_3y_end,
            "recall@",
            "100_abs",
            0.65,
            0.65,
            0.0,
            0.65,
            0.0,
        ),
        (
            spiky_grp,
            spiky_2y_end,
            "recall@",
            "100_abs",
            0.55,
            0.55,
            0.0,
            0.55,
            0.0,
        ),
        (
            spiky_grp,
            spiky_1y_end,
            "recall@",
            "100_abs",
            0.45,
            0.45,
            0.0,
            0.45,
            0.0,
        ),
    ]
    for dist_row in distance_rows:
        engine.execute(
            "insert into dist_table values (%s, %s, %s, %s, %s, %s, %s, %s, %s)",
            dist_row,
        )
    return distance_table, model_groups



from datetime import datetime

import pytest
import testing.postgresql
from sqlalchemy.engine import create_engine

from triage.component.architect.entity_date_table_generators import EntityDateTableGenerator

from . import utils


def test_empty_output():
    """An empty cohort table eagerly produces an error.

    (Rather than allowing execution to proceed.)

    """
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, "events", [])
        table_generator = EntityDateTableGenerator(
            query="select entity_id from events where outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            entity_date_table_name="exp_hash_cohort",
        )

        with pytest.raises(ValueError):
            # Request time outside of available intervals
            table_generator.generate_entity_date_table([datetime(2015, 12, 31)])

        (cohort_count,) = engine.execute(
            f"""\
            select count(*) from {table_generator.entity_date_table_name}
        """
        ).first()

        assert cohort_count == 0

        engine.dispose()


def test_entity_date_table_generator_replace():
    input_data = [
        (1, datetime(2016, 1, 1), True),
        (1, datetime(2016, 4, 1), False),
        (1, datetime(2016, 3, 1), True),
        (2, datetime(2016, 1, 1), False),
        (2, datetime(2016, 1, 1), True),
        (3, datetime(2016, 1, 1), True),
        (5, datetime(2016, 3, 1), True),
        (5, datetime(2016, 4, 1), True),
    ]
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, "events", input_data)
        table_generator = EntityDateTableGenerator(
            query="select entity_id from events where outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            entity_date_table_name="exp_hash_entity_date",
            replace=True
        )
        as_of_dates = [
            datetime(2016, 1, 1),
            datetime(2016, 2, 1),
            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
            datetime(2016, 5, 1),
            datetime(2016, 6, 1),
        ]
        table_generator.generate_entity_date_table(as_of_dates)
        expected_output = [
            (1, datetime(2016, 2, 1), True),
            (1, datetime(2016, 3, 1), True),
            (1, datetime(2016, 4, 1), True),
            (1, datetime(2016, 5, 1), True),
            (1, datetime(2016, 6, 1), True),
            (2, datetime(2016, 2, 1), True),
            (2, datetime(2016, 3, 1), True),
            (2, datetime(2016, 4, 1), True),
            (2, datetime(2016, 5, 1), True),
            (2, datetime(2016, 6, 1), True),
            (3, datetime(2016, 2, 1), True),
            (3, datetime(2016, 3, 1), True),
            (3, datetime(2016, 4, 1), True),
            (3, datetime(2016, 5, 1), True),
            (3, datetime(2016, 6, 1), True),
            (5, datetime(2016, 4, 1), True),
            (5, datetime(2016, 5, 1), True),
            (5, datetime(2016, 6, 1), True),
        ]
        results = list(
            engine.execute(
                f"""
                select entity_id, as_of_date, active from {table_generator.entity_date_table_name}
                order by entity_id, as_of_date
            """
            )
        )
        assert results == expected_output
        utils.assert_index(engine, table_generator.entity_date_table_name, "entity_id")
        utils.assert_index(engine, table_generator.entity_date_table_name, "as_of_date")

        table_generator.generate_entity_date_table(as_of_dates)
        assert results == expected_output


def test_entity_date_table_generator_noreplace():
    input_data = [
        (1, datetime(2016, 1, 1), True),
        (1, datetime(2016, 4, 1), False),
        (1, datetime(2016, 3, 1), True),
        (2, datetime(2016, 1, 1), False),
        (2, datetime(2016, 1, 1), True),
        (3, datetime(2016, 1, 1), True),
        (5, datetime(2016, 3, 1), True),
        (5, datetime(2016, 4, 1), True),
    ]
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, "events", input_data)
        table_generator = EntityDateTableGenerator(
            query="select entity_id from events where outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            entity_date_table_name="exp_hash_entity_date",
            replace=False
        )

        # 1. generate a cohort for a subset of as-of-dates
        as_of_dates = [
            datetime(2016, 1, 1),
            datetime(2016, 2, 1),
            datetime(2016, 3, 1),
        ]
        table_generator.generate_entity_date_table(as_of_dates)
        expected_output = [
            (1, datetime(2016, 2, 1), True),
            (1, datetime(2016, 3, 1), True),
            (2, datetime(2016, 2, 1), True),
            (2, datetime(2016, 3, 1), True),
            (3, datetime(2016, 2, 1), True),
            (3, datetime(2016, 3, 1), True),
        ]
        results = list(
            engine.execute(
                f"""
                select entity_id, as_of_date, active from {table_generator.entity_date_table_name}
                order by entity_id, as_of_date
            """
            )
        )
        assert results == expected_output
        utils.assert_index(engine, table_generator.entity_date_table_name, "entity_id")
        utils.assert_index(engine, table_generator.entity_date_table_name, "as_of_date")

        table_generator.generate_entity_date_table(as_of_dates)
        assert results == expected_output

        # 2. generate a cohort for a different subset of as-of-dates,
        # actually including an overlap to make sure that it doesn't double-insert anything
        as_of_dates = [
            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
            datetime(2016, 5, 1),
            datetime(2016, 6, 1),
        ]
        table_generator.generate_entity_date_table(as_of_dates)
        expected_output = [
            (1, datetime(2016, 2, 1), True),
            (1, datetime(2016, 3, 1), True),
            (1, datetime(2016, 4, 1), True),
            (1, datetime(2016, 5, 1), True),
            (1, datetime(2016, 6, 1), True),
            (2, datetime(2016, 2, 1), True),
            (2, datetime(2016, 3, 1), True),
            (2, datetime(2016, 4, 1), True),
            (2, datetime(2016, 5, 1), True),
            (2, datetime(2016, 6, 1), True),
            (3, datetime(2016, 2, 1), True),
            (3, datetime(2016, 3, 1), True),
            (3, datetime(2016, 4, 1), True),
            (3, datetime(2016, 5, 1), True),
            (3, datetime(2016, 6, 1), True),
            (5, datetime(2016, 4, 1), True),
            (5, datetime(2016, 5, 1), True),
            (5, datetime(2016, 6, 1), True),
        ]
        results = list(
            engine.execute(
                f"""
                select entity_id, as_of_date, active from {table_generator.entity_date_table_name}
                order by entity_id, as_of_date
            """
            )
        )
        assert results == expected_output



import json
import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

DEFAULT_KEYS = [
    "label_timespan",
    "label_name",
    "as_of_date_frequency",
    "max_training_history",
    "state",
    "cohort_name",
    "feature_groups",
]


class ModelGrouper:
    """Assign a model group id to given model input based on default or custom configuration

    This class is a wrapper around the `get_model_group_id` stored procedure,
    which interfaces with the model_groups table provision a stable model group id .
    The role of this class is mainly to provide data conversion, sensible defaults, and
    an abstraction layer over the database.

    Args:
        model_group_keys (list) A list of matrix metadata keys to uniquely define a model group.'
            In addition, the non-matrix attributes 'class_path' and 'parameters', referring to
            classifier training arguments, can be sent.
    """

    def __init__(self, model_group_keys=()):
        self.model_group_keys = frozenset(model_group_keys)

    def _final_model_group_args(self, class_path, parameters, matrix_metadata):
        """Generates model grouping arguments based on input.

        Applies a set of default or custom grouping keys depending on the object's
        configuration.

        Formats output in the structure recognized by the get_model_group_id stored procedure:
        {
            'class_path: (string)
            'parameters': (dict)
            'feature_names': (list)
            'model_config': (dict)
        }

        Args:
        class_path (string): a full class path for the classifier
        parameters (dict): all hyperparameters to be passed to the classifier
        matrix_metadata(dict): key-value pairs describing the configuration that produced
            a matrix used for training

        Returns: (dict) a dictionary of arguments suitable for the 'get_model_group_id'
            stored procedure
        """
        # step 1: is there an override?
        if len(self.model_group_keys) > 0:
            final = {}
            keys = set(self.model_group_keys)
            if "class_path" in keys:
                final["class_path"] = class_path
                keys.remove("class_path")
            else:
                final["class_path"] = ""

            if "parameters" in keys:
                final["parameters"] = parameters
                keys.remove("parameters")
            else:
                final["parameters"] = {}

            if "feature_names" in keys:
                final["feature_names"] = matrix_metadata["feature_names"]
                keys.remove("feature_names")
            else:
                final["feature_names"] = []

            final["model_config"] = {}

            for model_group_key in keys:
                final["model_config"][model_group_key] = matrix_metadata[
                    model_group_key
                ]

            return final

        # step 2: if no override, apply defaults
        else:
            model_config = {}
            for model_group_key in DEFAULT_KEYS:
                model_config[model_group_key] = matrix_metadata[model_group_key]

            return dict(
                class_path=class_path,
                parameters=parameters,
                feature_names=matrix_metadata["feature_names"],
                model_config=model_config,
            )

    def get_model_group_id(self, class_path, parameters, matrix_metadata, db_engine):
        """
        Returns model group id using store procedure 'get_model_group_id' which will
        return the same value for models with the same class_path, parameters,
        features, and model_config

        Args:
            class_path (string) A full classpath to the model class
            parameters (dict) hyperparameters to give to the model constructor
            matrix_metadata (dict) stored metadata about the train matrix
            db_engine (sqlalchemy.engine) A database engine pointing to a database with
             a results.model_groups table and get_model_group_id stored procedure

        Returns: (int) a database id for the model group id
        """
        model_group_args = self._final_model_group_args(
            class_path, parameters, matrix_metadata
        )
        db_conn = db_engine.raw_connection()
        cur = db_conn.cursor()
        cur.execute(
            "SELECT EXISTS ( "
            "       SELECT * "
            "       FROM pg_catalog.pg_proc "
            "       WHERE proname = 'get_model_group_id' ) "
        )
        condition = cur.fetchone()

        if condition:
            query = (
                "SELECT get_model_group_id( "
                "            '{class_path}'::TEXT, "
                "            '{parameters}'::JSONB, "
                "             ARRAY{feature_names}::TEXT [] , "
                "            '{model_config}'::JSONB )".format(
                    class_path=model_group_args["class_path"],
                    parameters=json.dumps(model_group_args["parameters"]),
                    feature_names=list(model_group_args["feature_names"]),
                    model_config=json.dumps(
                        model_group_args["model_config"], sort_keys=True
                    ),
                )
            )
            logger.spam(f"Getting model group from query {query}")
            cur.execute(query)
            db_conn.commit()
            model_group_id = cur.fetchone()
            model_group_id = model_group_id[0]

        else:
            logger.warning("Could not found stored procedure public.model_group_id")
            model_group_id = None
        db_conn.close()

        return model_group_id



import datetime
import unittest

from triage.util.conf import convert_str_to_relativedelta, parse_delta_string


class test_convert_str_to_relativedelta(unittest.TestCase):
    def test_valid_input(self):
        date = datetime.datetime(2016, 1, 1, 0, 0)
        for test in (
            {
                "interval": "1 year",
                "addition_result": datetime.datetime(2017, 1, 1, 0, 0),
                "subtraction_result": datetime.datetime(2015, 1, 1, 0, 0),
            },
            {
                "interval": "2 months",
                "addition_result": datetime.datetime(2016, 3, 1, 0, 0),
                "subtraction_result": datetime.datetime(2015, 11, 1, 0, 0),
            },
            {
                "interval": "3 days",
                "addition_result": datetime.datetime(2016, 1, 4, 0, 0),
                "subtraction_result": datetime.datetime(2015, 12, 29, 0, 0),
            },
            {
                "interval": "3 seconds",
                "addition_result": datetime.datetime(2016, 1, 1, 0, 0, 3),
                "subtraction_result": datetime.datetime(2015, 12, 31, 23, 59, 57),
            },
            {
                "interval": "2year",
                "addition_result": datetime.datetime(2018, 1, 1, 0, 0),
                "subtraction_result": datetime.datetime(2014, 1, 1, 0, 0),
            },
            {
                "interval": "4 weeks",
                "addition_result": datetime.datetime(2016, 1, 29, 0, 0),
                "subtraction_result": datetime.datetime(2015, 12, 4, 0, 0),
            },
            {
                "interval": "5 hours",
                "addition_result": datetime.datetime(2016, 1, 1, 5, 0),
                "subtraction_result": datetime.datetime(2015, 12, 31, 19, 0),
            },
            {
                "interval": "10minutes",
                "addition_result": datetime.datetime(2016, 1, 1, 0, 10),
                "subtraction_result": datetime.datetime(2015, 12, 31, 23, 50),
            },
            {
                "interval": "1microsecond",
                "addition_result": datetime.datetime(2016, 1, 1, 0, 0, 0, 1),
                "subtraction_result": datetime.datetime(
                    2015, 12, 31, 23, 59, 59, 999999
                ),
            },
            {
                "interval": "5m",
                "addition_result": datetime.datetime(2016, 1, 1, 0, 5),
                "subtraction_result": datetime.datetime(2015, 12, 31, 23, 55),
            },
        ):
            delta = convert_str_to_relativedelta(test["interval"])
            assert date + delta == test["addition_result"]
            assert date - delta == test["subtraction_result"]

    def test_bad_input(self):
        for bad_delta_string in ("4 tacos", "3"):
            with self.assertRaises(ValueError):
                convert_str_to_relativedelta(bad_delta_string)


class test_parse_delta_string(unittest.TestCase):
    def test_valid_input(self):
        delta_strings = ["1 year", "2 months", "3 days", "3s", "2y", "10m"]
        expected_results = [
            ("year", 1),
            ("months", 2),
            ("days", 3),
            ("s", 3),
            ("y", 2),
            ("m", 10),
        ]
        for index, delta_string in enumerate(delta_strings):
            result = parse_delta_string(delta_string)
            assert result == expected_results[index]

    def test_invalid_input(self):
        delta_strings = ["one year", "one_year", "ay"]
        for delta_string in delta_strings:
            with self.assertRaises(ValueError):
                parse_delta_string(delta_string)



from triage.validation_primitives import string_is_tablesafe
from hypothesis import given, example
from hypothesis.strategies import text, characters


# test with a variety of strings based on letters and numbers auto-generated by hypothesis
# and also add a hardcoded example that includes underscores because those are fine
@given(text(alphabet=characters(whitelist_categories=('Ll', 'Nd')), min_size=1))
@example('a_valid_name')
def test_string_is_tablesafe(s):
    assert string_is_tablesafe(s)


# test with a variety of strings based on unsafe characters auto-generated by hypothesis
# and also add a hardcoded example that should be bad because it has spaces
@given(text(alphabet='/ "A'))
@example('spaces are not valid')
@example('Neither_are_CAPITAL_letters')
def test_string_is_not_tablesafe(s):
    assert not string_is_tablesafe(s)



"""Associate experiments with models and matrices

Revision ID: 38f37d013686
Revises: d0ac573eaf1a
Create Date: 2018-10-18 16:31:43.181779

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '38f37d013686'
down_revision = 'd0ac573eaf1a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('experiment_matrices',
    sa.Column('experiment_hash', sa.String(), nullable=False),
    sa.Column('matrix_uuid', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['experiment_hash'], ['model_metadata.experiments.experiment_hash'], ),
    sa.PrimaryKeyConstraint('experiment_hash', 'matrix_uuid'),
    schema='model_metadata'
    )
    op.create_table('experiment_models',
    sa.Column('experiment_hash', sa.String(), nullable=False),
    sa.Column('model_hash', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['experiment_hash'], ['model_metadata.experiments.experiment_hash'], ),
    sa.PrimaryKeyConstraint('experiment_hash', 'model_hash'),
    schema='model_metadata'
    )
    op.add_column('matrices', sa.Column('built_by_experiment', sa.String(), nullable=True), schema='model_metadata')
    op.create_foreign_key(None, 'matrices', 'experiments', ['built_by_experiment'], ['experiment_hash'], source_schema='model_metadata', referent_schema='model_metadata')

    op.alter_column('models', 'experiment_hash', new_column_name='built_by_experiment', schema='model_metadata')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('models', 'built_by_experiment', new_column_name='experiment_hash', schema='model_metadata')
    op.drop_constraint(None, 'matrices', schema='model_metadata', type_='foreignkey')
    op.drop_column('matrices', 'built_by_experiment', schema='model_metadata')
    op.drop_table('experiment_models', schema='model_metadata')
    op.drop_table('experiment_matrices', schema='model_metadata')
    # ### end Alembic commands ###



import inspect


def classpath(klass):
    """Return the full class path

    Args:
        klass (class): A class
    """
    return f"{klass.__module__}.{klass.__name__}"


def bind_kwargs(kallable, **kwargs):
    """Bind keyword arguments to a callable and return as a dictionary

    Args:
        callable (callable): any callable
        **kwargs: keyword arguments to bind

    Returns: (dict)
    """
    call_signature = inspect.signature(kallable).bind_partial(**kwargs).arguments
    if 'kwargs' in call_signature:
        passed_kwargs = call_signature['kwargs']
    else:
        passed_kwargs = call_signature
    return passed_kwargs
